#!/usr/bin/env python3
# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Functions to interact with the keyboard
Mainly, read keys entered and return a string (most
of the time ONE char, but may be longer if it's a keyboard
shortcut, like ^A, M-a or KEY_RESIZE)
"""

import curses
import curses.ascii
import logging
from typing import Callable

log = logging.getLogger(__name__)

# A callback that will handle the next key entered by the user. For
# example if the user presses Ctrl+j, we set a callbacks, and the
# next key pressed by the user will be passed to this callback
# instead of the normal process of executing global keyboard
# shortcuts or inserting text in the current output.  The callback
# is always reset to None afterwards (to resume the normal
# processing of keys)
continuation_keys_callback: Callable[[str], None] | None = None


def get_char_list(s: curses.window) -> list[str]:
    ret_list: list[str] = []
    while True:
        try:
            key = s.get_wch()
        except curses.error:
            # No input, this means a timeout occurs.
            return ret_list
        except ValueError:  # invalid input
            log.debug('Invalid character entered.')
            return ret_list
        # Set to non-blocking. We try to read more bytes. If there are no
        # more data to read, it will immediately timeout and return with the
        # data we have so far
        s.timeout(0)
        if isinstance(key, int):
            ret_list.append(curses.keyname(key).decode())
        else:
            if curses.ascii.isctrl(key):
                key = curses.unctrl(key).decode()
                # Here special cases for alt keys, where we get a ^[ and then a second char
                if key == '^[':
                    try:
                        part = s.get_wch()
                        if part == '[':
                            # CTRL+arrow and meta+arrow keys have a long format
                            part += s.get_wch() + s.get_wch() + s.get_wch() + s.get_wch()  # type: ignore
                    except curses.error:
                        pass
                    except ValueError:  # invalid input
                        log.debug('Invalid character entered.')
                    else:
                        key = 'M-%s' % part
                    # and an even more special case for keys like
                    # ctrl+arrows, where we get ^[, then [, then a third
                    # char.
                    if key == 'M-[':
                        try:
                            part = s.get_wch()
                        except curses.error:
                            pass
                        except ValueError:
                            log.debug('Invalid character entered.')
                        else:
                            key = '%s-%s' % (key, part)
            if key == '\x7f' or key == '\x08':
                key = '^?'
            elif key == '\r':
                key = '^M'
            ret_list.append(key)


class Keyboard:
    def __init__(self) -> None:
        self.escape = False

    def escape_next_key(self) -> None:
        """
        The next key pressed by the user should be escaped. e.g. if the user
        presses ^N, keyboard.get_user_input() will return ["^", "N"] instead
        of ["^N"]. This will display ^N in the input, instead of
        interpreting the key binding.
        """
        self.escape = True

    def get_user_input(self, s: curses.window) -> list[str]:
        """
        Returns a list of all the available characters to read (for example it
        may contain a whole text if there’s some lag, or the user pasted text,
        or the user types really really fast).  Also it can return None, meaning
        that it’s time to do some other checks (because this function is
        blocking, we need to get out of it every now and then even if nothing
        was entered).
        """
        # Disable the timeout
        s.timeout(-1)
        ret_list = get_char_list(s)
        if not ret_list:
            return ret_list
        if len(ret_list) != 1:
            if ret_list[-1] == '^M':
                ret_list.pop(-1)
            ret_list = [char if char != '^M' else '^J' for char in ret_list]
        if self.escape:
            # Modify the first char of the list into its escaped version (i.e one or more char)
            key = ret_list.pop(0)
            for char in key[::-1]:
                ret_list.insert(0, char)
            self.escape = False
        return ret_list


if __name__ == '__main__':
    import sys
    keyboard = Keyboard()
    s = curses.initscr()
    curses.noecho()
    curses.cbreak()
    s.keypad(True)
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, 2, -1)
    s.attron(curses.A_BOLD | curses.color_pair(1))
    s.addstr('Type Ctrl-c to close\n')
    s.attroff(curses.A_BOLD | curses.color_pair(1))

    pressed_chars = []
    while True:

        try:
            chars = keyboard.get_user_input(s)
            for char in chars if chars else '':
                s.addstr('%s ' % (char))
            pressed_chars.append(chars)

        except KeyboardInterrupt:
            break
    curses.echo()
    curses.cbreak()
    curses.curs_set(1)
    curses.endwin()
    for char_list in pressed_chars:
        if char_list:
            print(' '.join((char for char in char_list)), end=' ')
    print()
    sys.exit(0)



"""
Global commands which are to be linked to the Core class
"""

from urllib.parse import unquote
from typing import TYPE_CHECKING, cast, Any
from xml.etree import ElementTree as ET
import logging

from slixmpp import JID, InvalidJID
from slixmpp.types import PresenceTypes
from slixmpp.exceptions import XMPPError, IqError, IqTimeout
from slixmpp.xmlstream.xmlstream import NotConnectedError
from slixmpp.xmlstream.stanzabase import StanzaBase
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath

from poezio import common, config as config_module, tabs, multiuserchat as muc
from poezio.bookmarks import Bookmark
from poezio.config import config, DEFAULT_CONFIG
from poezio.contact import Contact
from poezio.decorators import deny_anonymous
from poezio.plugin import PluginConfig
from poezio.roster import roster
from poezio.theming import dump_tuple, get_theme
from poezio.decorators import command_args_parser, command_jid_from_context
from poezio.core.structs import Command, POSSIBLE_SHOW

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class CommandCore:
    def __init__(self, core: "Core") -> None:
        self.core = core

    @command_args_parser.ignored
    def rotate_rooms_left(self) -> None:
        self.core.rotate_rooms_left()

    @command_args_parser.ignored
    def rotate_rooms_right(self) -> None:
        self.core.rotate_rooms_right()

    @command_args_parser.quoted(0, 1)
    def help(self, args: list[str]) -> None:
        """
        /help [command_name]
        """
        if not args:
            color = dump_tuple(get_theme().COLOR_HELP_COMMANDS)
            acc = []
            buff = ['Global commands:']
            for name, command in self.core.commands.items():
                if isinstance(command, Command):
                    acc.append('  \x19%s}%s\x19o - %s' % (color, name,
                                                          command.short_desc))
                else:
                    acc.append('  \x19%s}%s\x19o' % (color, name))
            acc = sorted(acc)
            buff.extend(acc)
            acc = []
            buff.append('Tab-specific commands:')
            tab_commands = self.core.tabs.current_tab.commands
            for name, command in tab_commands.items():
                if isinstance(command, Command):
                    acc.append('  \x19%s}%s\x19o - %s' % (color, name,
                                                          command.short_desc))
                else:
                    acc.append('  \x19%s}%s\x19o' % (color, name))
            acc = sorted(acc)
            buff.extend(acc)

            msg = '\n'.join(buff)
            msg += "\nType /help <command_name> to know what each command does"
        else:
            command_str = args[0].lstrip('/').strip()

            tab_commands = self.core.tabs.current_tab.commands
            if command_str in tab_commands:
                tup = tab_commands[command_str]
            elif command_str in self.core.commands:
                tup = self.core.commands[command_str]
            else:
                self.core.information(f'Unknown command: {command_str}', 'Error')
                return
            if isinstance(tup, Command):
                msg = 'Usage: /%s %s\n' % (command_str, tup.usage)
                msg += tup.desc
            else:
                msg = tup[1]
        self.core.information(msg, 'Help')

    @command_args_parser.quoted(1)
    def runkey(self, args: list[str]) -> None:
        """
        /runkey <key>
        """

        def replace_line_breaks(key: str) -> str:
            "replace ^J with \n"
            if key == '^J':
                return '\n'
            return key

        if not args:
            return self.help('runkey')
        char = args[0]
        func = self.core.key_func.get(char, None)
        if func is not None:
            func()
        else:
            self.core.do_command(replace_line_breaks(char), False)

    @command_args_parser.quoted(1, 1, [None])
    def status(self, args: list[str]) -> None:
        """
        /status <status> [msg]
        """
        if not args:
            return self.help('status')

        if args[0] not in POSSIBLE_SHOW.keys():
            return self.help('status')

        show = POSSIBLE_SHOW[args[0]]
        msg = args[1]

        pres = self.core.xmpp.make_presence()
        if msg:
            pres['status'] = msg
        pres['type'] = show
        self.core.events.trigger('send_normal_presence', pres)
        pres.send()
        current = self.core.tabs.current_tab
        if isinstance(current, tabs.MucTab) and current.joined and show in ('away', 'xa'):
            current.send_chat_state('inactive')
        for tab in self.core.tabs:
            if isinstance(tab, tabs.MucTab) and tab.joined:
                muc.change_show(self.core.xmpp, tab.jid, tab.own_nick, show,
                                msg)
            if hasattr(tab, 'directed_presence'):
                del tab.directed_presence
        self.core.set_status(show or '', msg)
        if isinstance(current, tabs.MucTab) and current.joined and show not in ('away', 'xa'):
            current.send_chat_state('active')

    @command_args_parser.quoted(1, 2, [None, None])
    def presence(self, args: list[str] | None) -> None:
        """
        /presence <JID> [type] [status]
        """
        if not args:
            return self.help('presence')

        ptype: str | None
        jid_str, ptype, status = args[0], args[1], args[2]
        if jid_str == '.' and isinstance(self.core.tabs.current_tab, tabs.ChatTab):
            jid = self.core.tabs.current_tab.jid
        else:
            try:
                jid = JID(jid_str)
            except InvalidJID:
                self.core.information(f'Invalid JID: {jid_str}', 'Error')
                return
        final_ptype: PresenceTypes | None
        if ptype not in ('error', 'probe', 'subscribe', 'subscribed',
                         'unavailable', 'unsubscribe', 'unsubscribed'):
            final_ptype = None
        else:
            final_ptype = cast(PresenceTypes, ptype)
        try:
            pres = self.core.xmpp.make_presence(
                pto=jid, ptype=final_ptype, pstatus=status)
            self.core.events.trigger('send_normal_presence', pres)
            pres.send()
        except (XMPPError, NotConnectedError):
            self.core.information('Could not send directed presence', 'Error')
            log.debug(
                'Could not send directed presence to %s', jid, exc_info=True)
            return
        tab = self.core.tabs.by_name(str(jid))
        if tab and isinstance(tab, tabs.ChatTab):
            if ptype in ('xa', 'away'):
                tab.directed_presence = False
                chatstate = 'inactive'
            else:
                tab.directed_presence = True
                chatstate = 'active'
            if tab == self.core.tabs.current_tab:
                tab.send_chat_state(chatstate, True)
            if isinstance(tab, tabs.MucTab):
                for private in tab.privates:
                    private.directed_presence = tab.directed_presence
                if self.core.tabs.current_tab in tab.privates:
                    self.core.tabs.current_tab.send_chat_state(chatstate, True)

    @command_args_parser.quoted(1)
    def theme(self, args: list[str] | None = None) -> None:
        """/theme <theme name>"""
        if not args:
            return self.help('theme')
        self.set_cmd('theme %s' % (args[0], ))

    @command_args_parser.quoted(1)
    def win(self, args: list[str] | str | None) -> None:
        """
        /win <number or name>
        """
        if not args:
            return self.help('win')

        name = args[0]
        try:
            number = int(name)
        except ValueError:
            number = -1
            name = name.lower()
        if number != -1 and self.core.tabs.current_tab.nb == number:
            return
        prev_nb = self.core.previous_tab_nb
        self.core.previous_tab_nb = self.core.tabs.current_tab.nb
        if 0 <= number < len(self.core.tabs):
            if not self.core.tabs[number]:
                self.core.previous_tab_nb = prev_nb
                return
            self.core.tabs.set_current_index(number)
        else:
            match = self.core.tabs.find_match(name)
            if match is None:
                return
            self.core.tabs.set_current_tab(match)

    @command_args_parser.quoted(1)
    def wup(self, args: list[str] | None) -> None:
        """
        /wup <prefix of name>
        """
        if not args:
            return self.help('wup')

        prefix = args[0]
        _, match = self.core.tabs.find_by_unique_prefix(prefix)
        if match is None:
            return
        self.core.tabs.set_current_tab(match)

    @command_args_parser.quoted(2)
    def move_tab(self, args: list[str]) -> None:
        """
        /move_tab old_pos new_pos
        """
        if not args:
            return self.help('move_tab')

        current_tab = self.core.tabs.current_tab
        if args[0] == '.':
            args[0] = str(current_tab.nb)
        if args[1] == '.':
            args[1] = str(current_tab.nb)

        def get_nb_from_value(value: str) -> int | None:
            "parse the cmdline to guess the tab the users wants"
            ref = None
            try:
                ref = int(value)
            except ValueError:
                old_tab = None
                for tab in self.core.tabs:
                    if not old_tab and value == tab.name:
                        old_tab = tab
                if not old_tab:
                    self.core.information("Tab %s does not exist" % args[0],
                                          "Error")
                    return None
                ref = old_tab.nb
            return ref

        old = get_nb_from_value(args[0])
        new = get_nb_from_value(args[1])
        if new is None or old is None:
            self.core.information('Unable to move the tab.', 'Info')
            return
        result = self.core.insert_tab(old, new)
        if not result:
            self.core.information('Unable to move the tab.', 'Info')
        self.core.refresh_window()

    @command_args_parser.quoted(0, 1)
    def list_cmd(self, args: list[str]) -> None:
        """
        /list [server]
        Opens a MucListTab containing the list of the room in the specified server
        """
        if not args:
            return self.help('list')
        elif args:
            try:
                jid = JID(args[0])
            except InvalidJID:
                self.core.information('Invalid server %r' % args[0], 'Error')
                return
        else:
            if not isinstance(self.core.tabs.current_tab, tabs.MucTab):
                self.core.information('Please provide a server',
                                      'Error')
                return
            jid = self.core.tabs.current_tab.jid
        if jid is None or not jid.domain:
            return
        self.core.register_task(
            self._list_async(jid)
        )

    async def _list_async(self, jid: JID) -> None:
        jid = JID(jid.domain)
        list_tab = tabs.MucListTab(self.core, jid)
        self.core.add_tab(list_tab, True)
        iq = await self.core.xmpp.plugin['xep_0030'].get_items(jid=jid)
        list_tab.on_muc_list_item_received(iq)

    @command_args_parser.quoted(1)
    async def version(self, args: list[str] | str | None) -> None:
        """
        /version <jid>
        """
        if not args:
            self.help('version')
            return
        try:
            jid = JID(args[0])
        except InvalidJID:
            self.core.information(
                'Invalid JID for /version: %s' % args[0],
                'Error'
            )
            return
        if jid.resource or jid not in roster or not roster.contacts[jid].resources:
            iq = await self.core.xmpp.plugin['xep_0092'].get_version(jid)
            self.core.handler.on_version_result(iq)
        elif jid in roster:
            for resource in roster.contacts[jid].resources:
                iq = await self.core.xmpp.plugin['xep_0092'].get_version(
                    JID(resource.jid)
                )
                self.core.handler.on_version_result(iq)

    def _empty_join(self) -> tuple[str | None, str | None]:
        tab = self.core.tabs.current_tab
        if not isinstance(tab, (tabs.MucTab, tabs.PrivateTab)):
            return (None, None)
        room = tab.jid.bare
        nick = tab.own_nick
        return (room, nick)

    def _parse_join_jid(self, jid_string: str
                        ) -> tuple[str, str | None] | tuple[None, None]:
        # we try to join a server directly
        server_root = False
        if jid_string.startswith('xmpp:') and jid_string.endswith('?join'):
            jid_string = unquote(jid_string[5:-5])
        try:
            if jid_string.startswith('@'):
                server_root = True
                info = JID(jid_string[1:])
            else:
                info = JID(jid_string)
                server_root = False
        except InvalidJID:
            info = JID('')

        set_nick: str | None = ''
        if len(jid_string) > 1 and jid_string.startswith('/'):
            set_nick = jid_string[1:]
        elif info.resource:
            set_nick = info.resource

        # happens with /join /nickname, which is OK
        if info.bare == '':
            tab = self.core.tabs.current_tab
            if not isinstance(tab, tabs.MucTab):
                return (None, None)
            else:
                room = tab.jid.bare
                if not set_nick:
                    set_nick = tab.own_nick
        else:
            room = info.bare
            # no server is provided, like "/join hello":
            # use the server of the current room if available
            # check if the current room's name has a server
            if room.find('@') == -1 and not server_root:
                tab = self.core.tabs.current_tab
                if isinstance(tab, tabs.MucTab) and tab.jid.domain:
                    room += '@%s' % tab.jid.domain
        return (room, set_nick)

    @command_args_parser.quoted(0, 2)
    async def join(self, args: list[str] | str) -> None:
        """
        /join [room][/nick] [password]
        """
        if not args:
            room, nick = self._empty_join()
        else:
            room, nick = self._parse_join_jid(args[0])
        if not room:
            return  # nothing was parsed

        room = room.lower()

        # Has the nick been picked from configuration, or chosen
        # explicitly when joining?
        config_nick = False
        if not nick:
            config_nick = True
            nick = self.core.own_nick

        # a password is provided
        if len(args) == 2:
            password = args[1]
        else:
            password = config.get_by_tabname('password', JID(room), fallback=False)

        if room in self.core.pending_invites:
            del self.core.pending_invites[room]

        tab = self.core.tabs.by_name_and_class(room, tabs.MucTab)
        # New tab
        if tab is None:
            tab = self.core.open_new_room(JID(room), nick, password=password)
            tab.join()
        else:
            self.core.focus_tab(tab)
            if tab.own_nick == nick and tab.joined:
                self.core.information('/join: Nothing to do.', 'Info')
            else:
                tab.command_part('')
                tab.own_nick = nick
                tab.password = password
                tab.join()

        if config.getbool('synchronise_open_rooms') and room not in self.core.bookmarks:
            method = 'remote' if config.getbool(
                'use_remote_bookmarks') else 'local'
            await self._add_bookmark(
                room=room,
                nick=nick if not config_nick else None,
                autojoin=True,
                password=password,
                method=method,
            )

        if tab == self.core.tabs.current_tab:
            tab.refresh()
            self.core.doupdate()

    @command_args_parser.quoted(0, 2)
    def bookmark_local(self, args: list[str]) -> None:
        """
        /bookmark_local [room][/nick] [password]
        """
        tab = self.core.tabs.current_tab
        nick: str | None
        if not args:
            if not isinstance(tab, tabs.MucTab):
                return
            room = tab.jid.bare
            nick = tab.own_nick
            password = tab.password
        else:
            tmp_room, tmp_nick = self._parse_join_jid(args[0] if args else '')
            if tmp_room is None:
                return
            room, nick = tmp_room, tmp_nick
            password = args[1] if len(args) > 1 else None

        self.core.register_task(
            self._add_bookmark(
                room=str(room),
                nick=nick,
                autojoin=True,
                password=password,
                method='local',
            )
        )

    @command_args_parser.quoted(0, 3)
    def bookmark(self, args: list[str]) -> None:
        """
        /bookmark [room][/nick] [autojoin] [password]
        """
        tab = self.core.tabs.current_tab
        nick: str | None
        if not args:
            if not isinstance(tab, tabs.MucTab):
                return
            room = tab.jid.bare
            nick = tab.own_nick
            password = tab.password
        else:
            tmp_room, tmp_nick = self._parse_join_jid(args[0])
            if tmp_room is None:
                return
            room, nick = tmp_room, tmp_nick
            password = args[2] if len(args) > 2 else None

        method = 'remote' if config.getbool('use_remote_bookmarks') else 'local'
        autojoin = (method == 'local' or
                    (len(args) > 1 and args[1].lower() == 'true'))

        self.core.register_task(
            self._add_bookmark(str(room), nick, autojoin, password, method)
        )

    async def _add_bookmark(
        self,
        room: str,
        nick: str | None,
        autojoin: bool,
        password: str | None,
        method: str,
    ) -> None:
        '''
        Adds a bookmark.

        Args:
            room: room Jid.
            nick: optional nick. Will always be added to the bookmark if
                specified. This takes precedence over tab.own_nick which takes
                precedence over core.own_nick (global config).
            autojoin: set the bookmark to join automatically.
            password: room password.
            method: 'local' or 'remote'.
        '''

        if room == '*':
            return await self._add_wildcard_bookmarks(method)

        # Once we found which room to bookmark, find corresponding tab if it
        # exists and fill nickname if none was specified and not default.
        tab = self.core.tabs.by_name_and_class(room, tabs.MucTab)
        if tab and isinstance(tab, tabs.MucTab) and \
           tab.joined and tab.own_nick != self.core.own_nick:
            nick = nick or tab.own_nick

        # Validate / Normalize
        try:
            if not nick:
                jid = JID(room)
            else:
                jid = JID('{}/{}'.format(room, nick))
            room = jid.bare
            nick = jid.resource or None
        except InvalidJID:
            self.core.information(f'Invalid address for bookmark: {room}/{nick}', 'Error')
            return

        bookmark = self.core.bookmarks[room]
        if bookmark is None:
            bookmark = Bookmark(room)
            self.core.bookmarks.append(bookmark)
        bookmark.method = method
        bookmark.autojoin = autojoin
        if nick:
            bookmark.nick = nick
        if password:
            bookmark.password = password

        self.core.bookmarks.save_local()
        try:
            result = await self.core.bookmarks.save_remote(
                self.core.xmpp,
            )
            self.core.handler.on_bookmark_result(result)
        except (IqError, IqTimeout) as iq:
            self.core.handler.on_bookmark_result(iq)

    async def _add_wildcard_bookmarks(self, method: str) -> None:
        new_bookmarks = []
        for tab in self.core.get_tabs(tabs.MucTab):
            bookmark = self.core.bookmarks[tab.jid.bare]
            if not bookmark:
                bookmark = Bookmark(tab.jid.bare, autojoin=True, method=method)
                new_bookmarks.append(bookmark)
            else:
                bookmark.method = method
                new_bookmarks.append(bookmark)
                self.core.bookmarks.remove(bookmark)
        new_bookmarks.extend(self.core.bookmarks.bookmarks)
        self.core.bookmarks.set(new_bookmarks)
        self.core.bookmarks.save_local()
        try:
            iq = await self.core.bookmarks.save_remote(self.core.xmpp)
            self.core.handler.on_bookmark_result(iq)
        except IqError as iq:
            self.core.handler.on_bookmark_result(iq)

    @command_args_parser.ignored
    def bookmarks(self) -> None:
        """/bookmarks"""
        tab = self.core.tabs.by_name_and_class('Bookmarks', tabs.BookmarksTab)
        if tab:
            self.core.tabs.set_current_tab(tab)
        else:
            tab = tabs.BookmarksTab(self.core, self.core.bookmarks)
            self.core.tabs.append(tab)
            self.core.tabs.set_current_tab(tab)

    @command_args_parser.quoted(0, 1)
    def remove_bookmark(self, args: list[str]) -> None:
        """/remove_bookmark [jid]"""
        jid = None
        if not args:
            tab = self.core.tabs.current_tab
            if isinstance(tab, tabs.MucTab):
                jid = tab.jid.bare
        else:
            jid = args[0]

        if jid:
            self.core.register_task(
                self._remove_bookmark_routine(jid)
            )

    async def _remove_bookmark_routine(self, jid: str) -> None:
        """Asynchronously remove a bookmark"""
        if self.core.bookmarks[jid]:
            self.core.bookmarks.remove(jid)
            try:
                await self.core.bookmarks.save(self.core.xmpp)
                self.core.information('Bookmark deleted', 'Info')
            except (IqError, IqTimeout):
                self.core.information('Error while deleting the bookmark',
                                      'Error')
        else:
            self.core.information('No bookmark to remove', 'Info')

    @deny_anonymous
    @command_args_parser.quoted(0, 1)
    def accept(self, args: list[str]) -> None:
        """
        Accept a JID. Authorize it AND subscribe to it
        """
        if not args:
            tab = self.core.tabs.current_tab
            RosterInfoTab = tabs.RosterInfoTab
            if not isinstance(tab, RosterInfoTab):
                self.core.information('No JID specified', 'Error')
                return
            else:
                item = tab.selected_row
                if isinstance(item, Contact):
                    bare_jid = item.bare_jid.bare
                else:
                    self.core.information('No subscription to accept', 'Warning')
                    return
        else:
            try:
                bare_jid = JID(args[0]).bare
            except InvalidJID:
                self.core.information('Invalid JID for /accept: %s' % args[0], 'Error')
                return
        jid = JID(bare_jid)
        nodepart = jid.user
        # crappy transports putting resources inside the node part
        if '\\2f' in nodepart:
            jid.user = nodepart.split('\\2f')[0]
        contact = roster[jid]
        if contact is None:
            self.core.information('No subscription to accept', 'Warning')
            return
        contact.pending_in = False
        roster.modified()
        self.core.xmpp.send_presence(pto=jid, ptype='subscribed')
        self.core.xmpp.client_roster.send_last_presence()
        if contact.subscription in ('from',
                                    'none') and not contact.pending_out:
            self.core.xmpp.send_presence(
                pto=jid, ptype='subscribe', pnick=self.core.own_nick)
        self.core.information('%s is now authorized' % jid, 'Roster')

    @deny_anonymous
    @command_args_parser.quoted(1)
    def add(self, args: list[str]) -> None:
        """
        Add the specified JID to the roster, and automatically
        accept the reverse subscription
        """
        if not args:
            tab = self.core.tabs.current_tab
            ConversationTab = tabs.ConversationTab
            if isinstance(tab, ConversationTab):
                jid = tab.general_jid
                if jid in roster and roster.contacts[jid].subscription in ('to', 'both'):
                    self.core.information('Already subscribed.', 'Roster')
                    return
                roster.add(jid)
                roster.modified()
                self.core.information('%s was added to the roster' % jid, 'Roster')
            else:
                self.core.information('No JID specified', 'Error')
        try:
            jid = JID(JID(args[0]).bare)
        except InvalidJID:
            self.core.information('Invalid JID for /add: %s' % args[0], 'Error')
            return
        if jid in roster and roster.contacts[jid].subscription in ('to', 'both'):
            self.core.information('Already subscribed.', 'Roster')
            return
        roster.add(jid)
        roster.modified()
        self.core.information('%s was added to the roster' % jid, 'Roster')

    @deny_anonymous
    @command_args_parser.quoted(0, 1)
    def deny(self, args: list[str]) -> None:
        """
        /deny [jid]
        Denies a JID from our roster
        """
        jid = None
        if not args:
            tab = self.core.tabs.current_tab
            if isinstance(tab, tabs.RosterInfoTab):
                item = tab.roster_win.selected_row
                if isinstance(item, Contact):
                    jid = item.bare_jid
        else:
            try:
                jid = JID(JID(args[0]).bare)
            except InvalidJID:
                self.core.information('Invalid JID for /deny: %s' % args[0], 'Error')
                return
            if jid not in [jid for jid in roster.jids()]:
                jid = None
        if jid is None:
            self.core.information('No subscription to deny', 'Warning')
            return

        contact = roster[jid]
        if contact:
            contact.unauthorize()
            self.core.information('Subscription to %s was revoked' % jid,
                                  'Roster')

    @deny_anonymous
    @command_args_parser.quoted(0, 1)
    def remove(self, args: list[str]) -> None:
        """
        Remove the specified JID from the roster. i.e.: unsubscribe
        from its presence, and cancel its subscription to our.
        """
        jid = None
        if args:
            try:
                jid = JID(JID(args[0]).bare)
            except InvalidJID:
                self.core.information('Invalid JID for /remove: %s' % args[0], 'Error')
                return
        else:
            tab = self.core.tabs.current_tab
            if isinstance(tab, tabs.RosterInfoTab):
                item = tab.roster_win.selected_row
                if isinstance(item, Contact):
                    jid = item.bare_jid
        if jid is None:
            self.core.information('No roster item to remove', 'Error')
            return
        roster.remove(jid)
        del roster[jid]

    @command_args_parser.ignored
    def command_reconnect(self) -> None:
        """
        /reconnect
        """
        if self.core.xmpp.is_connected():
            self.core.disconnect(reconnect=True)
        else:
            self.core.xmpp.start()

    @command_args_parser.quoted(0, 3)
    def set_cmd(self, args: list[str] | str) -> None:
        """
        /set [module|][section] <option> [value]
        """
        info = ("", "")
        if len(args) == 3 and args[1] == '=':
            args = [args[0], args[2]]
        if not args:
            config_dict = config.to_dict()
            lines = []
            theme = get_theme()
            for section_name, section_dict in config_dict.items():
                lines.append(
                    '\x19%(section_col)s}[%(section)s]\x19o' % {
                        'section': section_name,
                        'section_col': dump_tuple(
                            theme.COLOR_INFORMATION_TEXT),
                    })
                for option_name, option_value in section_dict.items():
                    if (isinstance(option_name, str) and
                            'password' in option_name and
                            'eval_password' not in option_name):
                        option_value = '********'
                    lines.append(
                        '%s\x19%s}=\x19o%s' %
                        (option_name, dump_tuple(
                            theme.COLOR_REVISIONS_MESSAGE), option_value))
            info = ('Current  options:\n%s' % '\n'.join(lines), 'Info')
        elif len(args) == 1:
            option = args[0]
            if '=' in option:
                args = option.split('=', 1)
            else:
                value = config.get(option)
                if (isinstance(option, str) and
                        'password' in option and
                        'eval_password' not in option and
                        value is not None):
                    value = '********'
                info = ('%s=%s' % (option, value), 'Info')
        if len(args) == 2:
            if '|' in args[0]:
                plugin_name, section = args[0].split('|')[:2]
                if not section:
                    section = plugin_name
                option = args[1]
                if plugin_name not in self.core.plugin_manager.plugins:
                    file_name = self.core.plugin_manager.plugins_conf_dir / (
                        plugin_name + '.cfg')
                    plugin_config = PluginConfig(file_name, plugin_name)
                else:
                    plugin = self.core.plugin_manager.plugins[plugin_name]
                    if plugin:
                        plugin_config = plugin.config
                    else:
                        return
                value = plugin_config.get(option, default='', section=section)
                info = ('%s=%s' % (option, value), 'Info')
            else:
                possible_section = args[0]
                if (not config.has_option(section='Poezio', option=possible_section)
                        and config.has_section(possible_section)):
                    section = possible_section
                    option = args[1]
                    value = config.getstr(option, section=section)
                    info = ('%s=%s' % (option, value), 'Info')
                else:
                    option = args[0]
                    value = args[1]
                    info = config.set_and_save(option, value)
                    self.core.trigger_configuration_change(option, value)
        elif len(args) == 3:
            if '|' in args[0]:
                plugin_name, section = args[0].split('|')[:2]
                if not section:
                    section = plugin_name
                option = args[1]
                value = args[2]
                if plugin_name not in self.core.plugin_manager.plugins:
                    file_name = self.core.plugin_manager.plugins_conf_dir / (
                        plugin_name + '.cfg')
                    plugin_config = PluginConfig(file_name, plugin_name)
                else:
                    plugin = self.core.plugin_manager.plugins[plugin_name]
                    if plugin is not None:
                        plugin_config = plugin.config
                    else:
                        return
                info = plugin_config.set_and_save(option, value, section)
            else:
                current_tab = self.core.tabs.current_tab
                if args[0] == '.' and isinstance(current_tab, tabs.ChatTab):
                    name = current_tab.jid.bare
                    if not name:
                        self.core.information(
                            'Invalid tab to use the "." argument.', 'Error')
                        return
                    section = name
                else:
                    section = args[0]
                option = args[1]
                value = args[2]
                info = config.set_and_save(option, value, section)
                self.core.trigger_configuration_change(option, value, section=section)
        elif len(args) > 3:
            return self.help('set')
        self.core.information(*info)

    @command_args_parser.quoted(1, 2)
    def set_default(self, args: list[str] | None) -> None:
        """
        /set_default [section] <option>
        """
        if not args:
            return self.help('set_default')
        if len(args) == 1:
            option = args[0]
            section = 'Poezio'
        elif len(args) == 2:
            section = args[0]
            option = args[1]
        else:
            return self.help('set_default')

        default_config = DEFAULT_CONFIG.get(section, {})
        if option not in default_config:
            info = ("Option %s has no default value" % (option), "Error")
            self.core.information(*info)
            return
        self.set_cmd('%s %s %s' % (section, option, default_config[option]))

    @command_args_parser.quoted(1)
    def toggle(self, args: list[str] | None) -> None:
        """
        /toggle <option>
        shortcut for /set <option> toggle
        """
        if not args:
            return self.help('toggle')

        if args[0]:
            self.set_cmd('%s toggle' % args[0])

    @command_args_parser.quoted(1, 1)
    def server_cycle(self, args: list[str] | None) -> None:
        """
        Do a /cycle on each room of the given server.
        If none, do it on the server of the current tab
        """
        tab = self.core.tabs.current_tab
        message = ""
        if args:
            try:
                domain = JID(args[0]).domain
            except InvalidJID:
                self.core.information(
                    "Invalid server domain: %s" % args[0],
                    "Error"
                )
                return
            if len(args) == 2:
                message = args[1]
        else:
            if isinstance(tab, tabs.MucTab):
                domain = tab.jid.domain
            else:
                self.core.information("No server specified", "Error")
                return
        for tab in self.core.get_tabs(tabs.MucTab):
            if tab.jid.domain == domain:
                tab.leave_room(message)
                tab.join()

    @command_args_parser.quoted(1)
    async def last_activity(self, args: list[str] | str | None) -> None:
        """
        /last_activity <jid>
        """

        if not args:
            self.help('last_activity')
            return
        try:
            jid = JID(args[0])
        except InvalidJID:
            self.core.information(
                'Invalid JID for /last_activity: %s' % args[0],
                'Error',
            )
            return

        try:
            iq = await self.core.xmpp.plugin['xep_0012'].get_last_activity(jid)
        except IqError as error:
            if error.etype == 'auth':
                msg = 'You are not allowed to see the activity of %s' % jid
            else:
                msg = 'Error retrieving the activity of %s: %s' % (jid, error)
            self.core.information(msg, 'Error')
            return
        except IqTimeout:
            self.core.information(
                'Timeout while retrieving the last activity of %s' % jid,
                'Error',
            )
            return

        seconds = iq['last_activity']['seconds']
        status = iq['last_activity']['status']
        from_ = iq['from']
        if not from_.user:
            msg = 'The uptime of %s is %s.' % (
                from_, common.parse_secs_to_str(seconds))
        else:
            msg = 'The last activity of %s was %s ago%s' % (
                from_, common.parse_secs_to_str(seconds),
                (' and their last status was %s' % status)
                if status else '')
        self.core.information(msg, 'Info')

    @command_args_parser.quoted(2, 1, [None])
    async def invite(self, args: list[str] | str | None) -> None:
        """/invite <to> <room> [reason]"""

        if not args:
            self.help('invite')
            return

        reason = args[2]
        try:
            to = JID(args[0])
        except InvalidJID:
            self.core.information('Invalid JID specified for invite: %s' % args[0], 'Error')
            return None
        try:
            room = JID(JID(args[1]).bare)
        except InvalidJID:
            self.core.information('Invalid room JID specified to invite: %s' % args[1], 'Error')
            return
        result = await self.core.invite(to, room, reason=reason)
        if result:
            self.core.information('Invited %s to %s' % (to.bare, room), 'Info')

    @command_args_parser.quoted(1, 0)
    def impromptu(self, args: list[str] | str | None) -> None:
        """/impromptu <jid> [<jid> ...]"""

        if not args:
            self.help('impromptu')
            return

        jids = set()
        current_tab = self.core.tabs.current_tab
        if isinstance(current_tab, tabs.ConversationTab):
            jids.add(current_tab.general_jid)

        for jid in common.shell_split(' '.join(args)):
            try:
                bare = JID(jid).bare
            except InvalidJID:
                self.core.information('Invalid JID for /impromptu: %s' % args[0], 'Error')
                return
            jids.add(JID(bare))

        self.core.register_task(self.core.impromptu(jids))

    @command_args_parser.quoted(1, 1, [''])
    def decline(self, args: list[str] | None) -> None:
        """/decline <room@server.tld> [reason]"""
        if not args:
            self.help('decline')
            return
        try:
            jid = JID(args[0])
        except InvalidJID:
            self.core.information('Invalid JID for /decline: %s' % args[0], 'Error')
            return
        if jid.bare not in self.core.pending_invites:
            return
        reason = args[1]
        del self.core.pending_invites[jid.bare]
        self.core.xmpp.plugin['xep_0045'].decline_invite(
            jid.bare, self.core.pending_invites[jid.bare], reason)

    @command_args_parser.quoted(0, 1)
    @command_jid_from_context
    async def block(self, jid: JID | None) -> None:
        """Block a JID"""
        if not jid:
            self.help("block")
            return
        try:
            await self.core.xmpp.plugin['xep_0191'].block(jid)
        except (IqError, IqTimeout):
            self.core.information(
                'Could not block %s.' % jid, 'Error',
            )
        else:
            self.core.information('Blocked %s.' % jid, 'Info')

    @command_args_parser.quoted(0, 1)
    @command_jid_from_context
    async def unblock(self, jid: JID | None) -> None:
        """Unblock a JID"""
        if not jid:
            self.help("unblock")
            return
        try:
            await self.core.xmpp.plugin['xep_0191'].unblock(jid)
        except (IqError, IqTimeout):
            self.core.information('Could not unblock the contact.',
                                  'Error')
        else:
            self.core.information('Unblocked %s.' % jid, 'Info')

    @command_args_parser.quoted(0, 1)
    @command_jid_from_context
    async def report(self, jid: JID | None) -> None:
        """Reports a JID to the account server"""
        if not jid:
            self.help("report")
            return
        try:
            await self.core.xmpp.plugin['xep_0377'].report(jid)
        except (IqError, IqTimeout):
            self.core.information(f'Could not report {jid} to the server.', 'Error')
        else:
            self.core.information(f'Reported {jid}.', 'Info')

### Commands without a completion in this class ###

    @command_args_parser.ignored
    def invitations(self) -> None:
        """/invitations"""
        build = []
        for room, inviter in self.core.pending_invites.items():
            try:
                bare = JID(inviter).bare
            except InvalidJID:
                self.core.information(
                    f'Invalid JID found in /invitations: {inviter}',
                    'Error'
                )
            build.append(f'{room} by {bare}')
        if build:
            message = 'You are invited to the following rooms:\n' + ','.join(build)
        else:
            message = 'You do not have any pending invitations.'
        self.core.information(message, 'Info')

    @command_args_parser.quoted(0, 1, [None])
    def quit(self, args: list[str]) -> None:
        """
        /quit [message]
        """
        if not self.core.xmpp.is_connected():
            self.core.exit()
            return

        msg = args[0]
        self.core.save_config()
        self.core.plugin_manager.disable_plugins()
        self.core.xmpp.add_event_handler(
            "disconnected", self.core.exit, disposable=True)
        self.core.disconnect(msg)

    @command_args_parser.quoted(0, 3, ['', '', ''])
    def destroy_room(self, args: list[str]) -> None:
        """
        /destroy_room [JID [reason [alternative room JID]]]
        """
        async def do_destroy(room: JID, reason: str, altroom: JID | None) -> None:
            try:
                await self.core.xmpp['xep_0045'].destroy(room, reason, altroom)
            except (IqError, IqTimeout) as e:
                self.core.information('Unable to destroy room %s: %s' % (room, e), 'Info')
            else:
                self.core.information('Room %s destroyed' % room, 'Info')

        room: JID | None
        if not args[0] and isinstance(self.core.tabs.current_tab, tabs.MucTab):
            room = self.core.tabs.current_tab.general_jid
        else:
            try:
                room = JID(args[0])
            except InvalidJID:
                room = None
            else:
                if room.resource:
                    room = None

        if room is None:
            self.core.information('Invalid room JID: "%s"' % args[0], 'Error')
            return

        reason = args[1]
        altroom = None
        if args[2]:
            try:
                altroom = JID(args[2])
            except InvalidJID:
                self.core.information('Invalid alternative room JID: "%s"' % args[2], 'Error')
                return

        self.core.register_task(do_destroy(room, reason, altroom))

    @command_args_parser.quoted(1, 1, [''])
    def bind(self, args: list[str]) -> None:
        """
        Bind a key.
        """
        if not args:
            return self.help('bind')

        if not config.silent_set(args[0], args[1], section='bindings'):
            self.core.information('Unable to write in the config file',
                                  'Error')

        if args[1]:
            self.core.information('%s is now bound to %s' % (args[0], args[1]),
                                  'Info')
        else:
            self.core.information(
                '%s is now reset to the default binding' % args[0], 'Info')

    @command_args_parser.raw
    def rawxml(self, args: str) -> None:
        """
        /rawxml <xml stanza>
        """

        if not args:
            return

        raw_stanza = args
        try:
            stanza = StanzaBase(self.core.xmpp, xml=ET.fromstring(raw_stanza))
            if stanza.xml.tag == 'iq' and stanza.xml.attrib.get('type') in (
                    'get', 'set'):
                iq_id = stanza.xml.attrib.get('id')
                if not iq_id:
                    iq_id = self.core.xmpp.new_id()
                    stanza['id'] = iq_id

                def iqfunc(iq: StanzaBase) -> None:
                    "handler for an iq reply"
                    self.core.information(str(iq), 'Iq')
                    self.core.xmpp.remove_handler('Iq %s' % iq_id)

                self.core.xmpp.register_handler(
                    Callback('Iq %s' % iq_id, StanzaPath('iq@id=%s' % iq_id),
                             iqfunc))
            stanza.send()
        except Exception:
            self.core.information('Could not send custom stanza', 'Error')
            log.debug(
                '/rawxml: Could not send custom stanza (%s)',
                repr(raw_stanza),
                exc_info=True)

    @command_args_parser.quoted(1, 256)
    def load(self, args: list[str]) -> None:
        """
        /load <plugin> [<otherplugin> …]
        # TODO: being able to load more than 256 plugins at once, hihi.
        """

        usage = '/load <plugin> [<otherplugin> …]'
        if not args:
            self.core.information(usage, 'Error')
            return

        for plugin in args:
            self.core.plugin_manager.load(plugin)

    @command_args_parser.quoted(1, 256)
    def unload(self, args: list[str]) -> None:
        """
        /unload <plugin> [<otherplugin> …]
        """

        usage = '/unload <plugin> [<otherplugin> …]'
        if not args:
            self.core.information(usage, 'Error')
            return

        for plugin in args:
            self.core.plugin_manager.unload(plugin)

    @command_args_parser.ignored
    def plugins(self) -> None:
        """
        /plugins
        """
        self.core.information(
            "Plugins currently in use: %s" % repr(
                list(self.core.plugin_manager.plugins.keys())), 'Info')

    @command_args_parser.quoted(1, 1)
    async def message(self, args: list[str] | None) -> None:
        """
        /message <jid> [message]
        """
        if not args:
            return self.help('message')
        try:
            jid = JID(args[0])
        except InvalidJID:
            self.core.information('Invalid JID for /message: %s' % args[0], 'Error')
            return
        if not jid.user and not jid.domain and not jid.resource:
            self.core.information('Invalid JID.', 'Error')
            return
        tab: tabs.ChatTab | None
        tab = self.core.get_conversation_by_jid(
            jid.full, False, fallback_barejid=False)
        muc = self.core.tabs.by_name_and_class(jid.bare, tabs.MucTab)
        if not tab and not muc:
            tab = self.core.open_conversation_window(JID(jid.full), focus=True)
        elif muc is not None:
            if jid.resource:
                tab = self.core.tabs.by_name_and_class(jid.full,
                                                       tabs.PrivateTab)
                if tab:
                    self.core.focus_tab(tab)
                else:
                    tab = self.core.open_private_window(jid.bare, jid.resource)
            else:
                tab = muc
        elif tab is not None:
            self.core.focus_tab(tab)
        if len(args) == 2 and isinstance(tab, tabs.ChatTab):
            await tab.command_say(args[1])

    @command_args_parser.ignored
    def xml_tab(self) -> None:
        """/xml_tab"""
        xml_tab = self.core.focus_tab_named('XMLTab', tabs.XMLTab)
        if not xml_tab:
            tab = tabs.XMLTab(self.core)
            self.core.add_tab(tab, True)
            self.core.xml_tab = tab

    @command_args_parser.quoted(1)
    async def adhoc(self, args: list[str]) -> None:
        if not args:
            return self.help('ad-hoc')
        try:
            jid = JID(args[0])
        except InvalidJID:
            self.core.information(
                'Invalid JID for ad-hoc command: %s' % args[0],
                'Error',
            )
            return
        list_tab = tabs.AdhocCommandsListTab(self.core, jid)
        self.core.add_tab(list_tab, True)
        iq = await self.core.xmpp.plugin['xep_0050'].get_commands(
            jid=jid,
            local=False
        )
        list_tab.on_list_received(iq)

    @command_args_parser.ignored
    def self_(self) -> None:
        """
        /self
        """
        status = self.core.get_status()
        show, message = status.show, status.message
        nick = self.core.own_nick
        jid = self.core.xmpp.boundjid.full
        info = ('Your JID is %s\nYour current status is "%s" (%s)'
                '\nYour default nickname is %s\nYou are running poezio %s' %
                (jid, message if message else '', show
                 if show else 'available', nick, self.core.custom_version))
        self.core.information(info, 'Info')

    @command_args_parser.ignored
    def reload(self) -> None:
        """
        /reload
        """
        self.core.reload_config()

    @command_args_parser.raw
    def debug(self, args: str) -> None:
        """/debug [filename]"""
        if not args.strip():
            config_module.setup_logging('')
            self.core.information('Debug logging disabled!', 'Info')
        elif args:
            config_module.setup_logging(args)
            self.core.information(f'Debug logging to {args} enabled!', 'Info')


def dumb_callback(*args: Any, **kwargs: Any) -> None:
    "mock callback"



"""
Test the config module
"""

import os
import tempfile
from pathlib import Path

import pytest

from poezio import config
from slixmpp import JID


@pytest.fixture(scope="module")
def config_obj():
    file_ = tempfile.NamedTemporaryFile(delete=False)
    conf = config.Config(file_name=Path(file_.name))
    yield conf
    del conf
    os.unlink(file_.name)


class TestConfigSimple(object):
    def test_get_set(self, config_obj):
        config_obj.set_and_save('test', value='coucou')
        config_obj.set_and_save('test2', value='true')
        assert config_obj.get('test') == 'coucou'
        assert config_obj.get('test2') == 'true'
        assert config_obj.get('toto') == ''

    def test_file_content(self, config_obj):
        with open(str(config_obj.file_name), 'r') as fd:
            data = fd.read()
        supposed_content = '[Poezio]\ntest = coucou\ntest2 = true\n'
        assert data == supposed_content

    def test_get_types(self, config_obj):

        config_obj.set_and_save('test_int', '99')
        config_obj.set_and_save('test_int_neg', '-1')
        config_obj.set_and_save('test_bool_t', 'true')
        config_obj.set_and_save('test_bool_f', 'false')
        config_obj.set_and_save('test_float', '1.5')

        assert config_obj.get('test_int', default=0) == 99
        assert config_obj.get('test_int_neg', default=0) == -1
        assert config_obj.get('test_bool_t', default=False) == True
        assert config_obj.get('test_bool_f', default=True) == False
        assert config_obj.get('test_float', default=1.0) == 1.5

    def test_remove(self, config_obj):
        with open(str(config_obj.file_name), 'r') as fd:
            data = fd.read()

        supposed_content = ('[Poezio]\ntest = coucou\ntest2 = true\n'
                            'test_int = 99\ntest_int_neg = -1\ntest_bool_t ='
                            ' true\ntest_bool_f = false\ntest_float = 1.5\n')

        assert data == supposed_content

        config_obj.remove_and_save('test_int')
        config_obj.remove_and_save('test_int_neg')
        config_obj.remove_and_save('test_bool_t')
        config_obj.remove_and_save('test_bool_f')
        config_obj.remove_and_save('test_float')

        with open(str(config_obj.file_name), 'r') as fd:
            data = fd.read()

        supposed_content = '[Poezio]\ntest = coucou\ntest2 = true\n'

        assert data == supposed_content


    def test_toggle(self, config_obj):
        config_obj.set_and_save('test2', value='toggle')
        assert config_obj.get('test2') == 'false'
        config_obj.set_and_save('test2', value='toggle')
        assert config_obj.get('test2') == 'true'

    def test_get_set_default(self, config_obj):
        assert config_obj.get('doesnotexist', 'toto@tata') == 'toto@tata'
        assert config_obj.get('doesnotexist2', '1234') == '1234'

class TestConfigSections(object):
    def test_set_section(self, config_obj):
        config_obj.set_and_save('option1', 'test', section='NotPoezio')
        config_obj.set_and_save('option2', 'test2', section='NotPoezio')

        assert config_obj.get('option1', section='NotPoezio') == 'test'
        assert config_obj.get('option2', section='NotPoezio') == 'test2'

    def test_file_content(self, config_obj):
        with open(str(config_obj.file_name), 'r') as fd:
            data = fd.read()
        supposed_content = ('[Poezio]\ntest = coucou\ntest2 = true\n'
                            '[NotPoezio]\noption1 = test\noption2 = test2\n')
        assert data == supposed_content

class TestTabNames(object):
    def test_get_tabname(self, config_obj):
        config_obj.set_and_save('test', value='value.toto@toto.com',
                                section='toto@toto.com')
        config_obj.set_and_save('test2', value='value2@toto.com',
                                section='@toto.com')

        assert config_obj.get_by_tabname('test', JID('toto@toto.com')) == 'value.toto@toto.com'
        assert config_obj.get_by_tabname('test2', JID('toto@toto.com')) == 'value2@toto.com'
        assert config_obj.get_by_tabname('test2', JID('toto@toto.com'), fallback=False) == 'value2@toto.com'
        assert config_obj.get_by_tabname('test2', JID('toto@toto.com'), fallback_server=False) == 'true'
        assert config_obj.get_by_tabname('test_int', JID('toto@toto.com'), fallback=False) == ''





import pytest
from datetime import datetime

from poezio.ui.types import BaseMessage, Message, XMLLog


def test_create_message():
    now = datetime.now()
    msg = Message(
        txt="coucou",
        nickname="toto",
    )
    assert now < msg.time < datetime.now()

    msg = Message(
        txt="coucou",
        nickname="toto",
        time=now,
    )
    assert msg.time == now


def test_message_offset_simple():
    msg = Message(
        txt="coucou",
        nickname="toto",
    )
    example = "10:10:10 toto> "
    assert msg.compute_offset(True, 10) == len(example)

    msg = Message(
        txt="coucou",
        nickname="toto",
        history=True,
    )
    example = "2019:09:01 10:10:10 toto> "
    assert msg.compute_offset(True, 10) == len(example)

def test_message_offset_no_nick():
    msg = Message(
        txt="coucou",
        nickname="",
    )
    example = "10:10:10 "
    assert msg.compute_offset(True, 10) == len(example)

def test_message_offset_ack():
    msg = Message(
        txt="coucou",
        nickname="toto",
        ack=1,
    )
    example = "10:10:10 V toto> "
    assert msg.compute_offset(True, 10) == len(example)

    msg = Message(
        txt="coucou",
        nickname="toto",
        ack=-1,
    )
    example = "10:10:10 X toto> "
    assert msg.compute_offset(True, 10) == len(example)


def test_message_offset_me():
    msg = Message(
        txt="/me coucou",
        nickname="toto",
    )
    example = "10:10:10 * toto "
    assert msg.compute_offset(True, 10) == len(example)


def test_message_offset_revisions():
    msg = Message(
        txt="coucou",
        nickname="toto",
        revisions=3,
    )
    example = "10:10:10 toto3> "
    assert msg.compute_offset(True, 10) == len(example)

    msg = Message(
        txt="coucou",
        nickname="toto",
        revisions=250,
    )
    example = "10:10:10 toto250> "
    assert msg.compute_offset(True, 10) == len(example)


def test_message_repr_works():
    msg1 = Message(
        txt="coucou",
        nickname="toto",
        revisions=250,
    )
    msg2 = Message(
        txt="coucou",
        nickname="toto",
        old_message=msg1
    )

    assert repr(msg2) is not None

def test_xmllog_offset():
    msg = XMLLog(
        txt='toto',
        incoming=True,
    )
    example = '10:10:10 IN  '
    assert msg.compute_offset(True, 10) == len(example)

    msg = XMLLog(
        txt='toto',
        incoming=False,
    )
    example = '10:10:10 OUT '
    assert msg.compute_offset(True, 10) == len(example)

def test_basemessage_offset():
    msg = BaseMessage(
        txt='coucou',
    )
    example = '10:10:10 '
    assert msg.compute_offset(True, 10) == len(example)



# Type hints for hsluv, which is a generated file so we do not want to
# modify it;
from typing import TypedDict


FloatTuple = list[float]


class Bound(TypedDict):
    slope: float
    intercept: float


def _get_bounds(l: float) -> list[Bound]:
    ...


def _distance_line_from_origin(line: Bound) -> float:
    ...


def _length_of_ray_until_intercept(theta: float, line: Bound) -> float:
    ...


def _max_safe_chroma_for_l(l: float) -> float:
    ...


def _max_chroma_for_lh(l: float, h: float) -> float:
    ...


def _dot_product(a: FloatTuple, b: FloatTuple) -> float:
    ...


def _from_linear(c: float) -> float:
    ...


def _to_linear(c: float) -> float:
    ...


def xyz_to_rgb(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def rgb_to_xyz(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def _y_yo_l(y: float) -> float:
    ...


def _l_to_y(l: float) -> float:
    ...


def xyz_to_luv(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def luv_to_xyz(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def lch_to_luv(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def hsluv_to_lch(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def lch_to_hsluv(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def hpluv_to_lch(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def lch_to_hpluv(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def rgb_to_hex(_hx_tuple: FloatTuple) -> str:
    ...


def hex_to_rgb(hex: str) -> FloatTuple:
    ...


def lch_to_rgb(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def rgb_to_lch(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def hsluv_to_rgb(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def rgb_to_hsluv(_hx_tuple: tuple[float, float, float]) -> FloatTuple:
    ...


def hpluv_to_rgb(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def rgb_to_hpluv(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def hsluv_to_hex(_hx_tuple: FloatTuple) -> str:
    ...


def hpluv_to_hex(_hx_tuple: FloatTuple) -> FloatTuple:
    ...


def hex_to_hsluv(s: str) -> FloatTuple:
    ...


def hex_to_hpluv(s: str) -> FloatTuple:
    ...



"""
Test the functions in the `xhtml` module
"""

import pytest
import xml
import poezio.xhtml
from poezio.xhtml import (poezio_colors_to_html, xhtml_to_poezio_colors,
                   _parse_css as parse_css, clean_text)


class ConfigShim:
    def __init__(self):
        self.value = True
    def get(self, *args, **kwargs):
        return self.value
    def getbool(self, *args, **kwargs):
        return self.value


config = ConfigShim()
poezio.xhtml.config = config

def test_clean_text():
    example_string = '\x191}Toto \x192,-1}titi\x19b Tata'
    assert clean_text(example_string) == 'Toto titi Tata'

    clean_string = 'toto titi tata'
    assert clean_text(clean_string) == clean_string

def test_poezio_colors_to_html():
    base = "<body xmlns='http://www.w3.org/1999/xhtml'><p>"
    end = "</p></body>"
    text = '\x191}coucou'
    assert poezio_colors_to_html(text) == base + '<span style="color: red;">coucou</span>' + end

    text = '\x19bcoucou\x19o toto \x194}titi'
    assert poezio_colors_to_html(text) == base + '<span style="font-weight: bold;">coucou</span> toto <span style="color: blue;">titi</span>' + end

    text = '\x19icoucou'
    assert poezio_colors_to_html(text) == base + '<span style="font-style: italic;">coucou</span>' + end

def test_xhtml_to_poezio_colors():
    start = b'<body xmlns="http://www.w3.org/1999/xhtml"><p>'
    end = b'</p></body>'
    xhtml = start + b'test' + end
    assert xhtml_to_poezio_colors(xhtml) == 'test'

    xhtml = start + b'<a href="http://perdu.com">salut</a>' + end
    assert xhtml_to_poezio_colors(xhtml) == '\x19usalut\x19o (http://perdu.com)'

    xhtml = start + b'<a href="http://perdu.com">http://perdu.com</a>' + end
    assert xhtml_to_poezio_colors(xhtml) == '\x19uhttp://perdu.com\x19o'

    xhtml = start + b'<span style="font-style: italic">Test</span>' + end
    assert xhtml_to_poezio_colors(xhtml) == '\x19iTest\x19o'

    xhtml = b'<div style="font-weight:bold">Allo <div style="color:red">test <div style="color: blue">test2</div></div></div>'
    assert xhtml_to_poezio_colors(xhtml, force=True) == '\x19bAllo \x19196}test \x1921}test2\x19o'

    xhtml = (b'<div style="color:blue"><div style="color:yellow">'
             b'<div style="color:blue">Allo <div style="color:red">'
             b'test <div style="color: blue">test2</div></div></div></div></div>')
    assert xhtml_to_poezio_colors(xhtml, force=True) == '\x1921}Allo \x19196}test \x1921}test2\x19o'

    with pytest.raises(xml.sax._exceptions.SAXParseException):
        xhtml_to_poezio_colors(b'<p>Invalid xml')

def test_xhtml_to_poezio_colors_disabled():
    config.value = False
    start = b'<body xmlns="http://www.w3.org/1999/xhtml"><p>'
    end = b'</p></body>'
    xhtml = start + b'test' + end
    assert xhtml_to_poezio_colors(xhtml) == 'test'

    xhtml = start + b'<a href="http://perdu.com">salut</a>' + end
    assert xhtml_to_poezio_colors(xhtml) == '\x19usalut\x19o (http://perdu.com)'

    xhtml = start + b'<a href="http://perdu.com">http://perdu.com</a>' + end
    assert xhtml_to_poezio_colors(xhtml) == '\x19uhttp://perdu.com\x19o'

    xhtml = b'<div style="font-weight:bold">Allo <div style="color:red">test <div style="color: blue">test2</div></div></div>'
    assert xhtml_to_poezio_colors(xhtml, force=True) == 'Allo test test2'

    xhtml = (b'<div style="color:blue"><div style="color:yellow">'
             b'<div style="color:blue">Allo <div style="color:red">'
             b'test <div style="color: blue">test2</div></div></div></div></div>')
    assert xhtml_to_poezio_colors(xhtml, force=True) == 'Allo test test2'

    config.value = True

def test_parse_css():
    example_css = 'text-decoration: underline; color: red;'
    assert parse_css(example_css) == '\x19u\x19196}'

    example_css = 'text-decoration: underline coucou color: red;'
    assert parse_css(example_css) == ''



"""
Usage
-----

This plugin defines three new global commands: :term:`/remind`,
:term:`/done`, and :term:`/tasks`.

.. glossary::

    /remind
        **Usage:** ``/remind <time> <todo>``

        This command will remind you to do ``todo`` every ``time``.

    /done
        **Usage:** ``/done <id>``
        Remove a reminder.

        The ``id`` is found using :term:`/tasks`.


    /tasks

        Print a list of the tasks, their ids, and their frequency, into the
        information buffer.

Time format
-----------

In seconds:

.. code-block:: none

    /remind 600 Work!

Will remind you to work every 10 minutes.

Defining the time in seconds is not really practical, so you can describe it
with days, hours, and minutes, in a time-string, e.g:

.. code-block:: python

    /remind 1h23m "Get up"

Will remind you to get up every 1 hour 23 minutes.


"""

from poezio.core.structs import Completion
from poezio.plugin import BasePlugin
from poezio.windows import Input
from poezio import timed_events
from poezio import common
import curses


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'remind',
            self.command_remind,
            usage='<seconds> <todo>',
            help='Remind you of <todo> every <time> seconds.',
            short='Remind you of a task',
            completion=self.completion_remind)
        self.api.add_command(
            'done',
            self.command_done,
            usage='<id>',
            help='Stop reminding you do the task identified by <id>.',
            short='Remove a task',
            completion=self.completion_done)
        self.api.add_command(
            'tasks',
            self.command_tasks,
            usage='',
            help='List all the current tasks and their ids.',
            short='List current tasks')
        self.tasks = {}
        self.count = 0

        for option in self.config.options(self.__module__):
            id_str, secs = option.split(',')
            id = int(id_str)
            if id > self.count:
                self.count = id
            value = self.config.get(option, '')
            self.tasks[id] = (int(secs), value)
            self.config.remove_section(self.__module__)
            self.config.add_section(self.__module__)
        if self.tasks:
            self.count += 1
            self.command_tasks('', nocommand=True)

    def command_remind(self, str_arg: str) -> None:
        args = common.shell_split(str_arg)
        if len(args) < 2:
            return
        time = common.parse_str_to_secs(args[0])
        if not time:
            return

        self.tasks[self.count] = (time, args[1])
        timed_event = timed_events.DelayedEvent(time, self.remind, self.count)
        self.api.add_timed_event(timed_event)
        self.api.information(
            'Task %s added: %s every %s.' % (self.count, args[1],
                                             common.parse_secs_to_str(time)),
            'Info')
        self.count += 1

    def completion_remind(self, the_input: Input) -> Completion | None:
        txt = the_input.get_text()
        args = common.shell_split(txt)
        n = len(args)
        if txt.endswith(' '):
            n += 1
        if n == 2:
            return Completion(the_input.auto_completion,
                              ["60", "5m", "15m", "30m", "1h", "10h", "1d"],
                              '')
        return None

    def completion_done(self, the_input: Input) -> Completion:
        return Completion(the_input.auto_completion,
                          ["%s" % key for key in self.tasks], '')

    def command_done(self, arg: str = "0") -> None:
        try:
            id_ = int(arg)
        except ValueError:
            return
        if id_ not in self.tasks:
            return

        self.api.information('Task %s: %s [DONE]' % (id_, self.tasks[id_][1]),
                             'Info')
        del self.tasks[id_]

    def command_tasks(self, arg: str, nocommand: bool = False) -> None:
        if nocommand:
            s = 'The following tasks were loaded:\n'
        else:
            s = 'The following tasks are active:\n'
        for key in self.tasks:
            s += 'Task %s: %s every %s.\n' % (key, repr(self.tasks[key][1]),
                                              common.parse_secs_to_str(
                                                  self.tasks[key][0]))
        if s:
            self.api.information(s, 'Info')

    def remind(self, id_: int = 0) -> None:
        if id_ not in self.tasks:
            return
        self.api.information('Task %s: %s' % (id_, self.tasks[id_][1]), 'Info')
        if self.config.get('beep', '') == 'true':
            curses.beep()
        timed_event = timed_events.DelayedEvent(self.tasks[id_][0],
                                                self.remind, id_)
        self.api.add_timed_event(timed_event)

    def cleanup(self) -> None:
        if self.tasks:
            self.config.remove_section(self.__module__)
            self.config.add_section(self.__module__)
            for task in self.tasks:
                self.config.set('%s,%s' % (task, self.tasks[task][0]),
                                self.tasks[task][1])
        self.config.write()



"""
This plugin adds several aliases, to shorten status changes.

Aliases
-------

.. glossary::
    :sorted:

    /afk
    /away
        Set your status to ``away``

    /dnd
    /busy
        Set your status to ``dnd``

    /available
        Set your status to ``available``

    /chat
        Set your status to ``chat``

    /xa
        Set your status to ``xa``

"""
from poezio.plugin import BasePlugin


class Plugin(BasePlugin):
    """
    Adds several convenient aliases to /status command
    """

    def init(self) -> None:
        for st in ('dnd', 'busy', 'afk', 'chat', 'xa', 'away', 'available'):
            def command_status(line: str, st: str = st) -> None:
                self.api.run_command('/status ' + st + ' "' + line + '"')

            self.api.add_command(
                st,
                command_status,
                usage='[status message]',
                short='Set your status as %s' % st,
                help='Set your status as %s' % st)



"""
Test the functions in the `common` module
"""

import time
import pytest
import datetime
from slixmpp import JID
from datetime import timedelta
from poezio.common import (_datetime_tuple as datetime_tuple, get_utc_time,
                           get_local_time, shell_split, _find_argument_quoted
                           as find_argument_quoted, _find_argument_unquoted as
                           find_argument_unquoted, parse_str_to_secs,
                           parse_secs_to_str, unique_prefix_of)

def test_utc_time():
    delta = timedelta(seconds=-3600)
    d = datetime.datetime.now()
    time.timezone = -3600; time.altzone = -3600
    assert get_utc_time(local_time=d) == d + delta

def test_local_time():
    delta = timedelta(seconds=-3600)
    d = datetime.datetime.now()
    time.timezone = -3600
    time.altzone = -3600
    assert get_local_time(d) == d - delta

def test_shell_split():
    assert shell_split('"sdf 1" "toto 2"') == ['sdf 1', 'toto 2']
    assert shell_split('toto "titi"') == ['toto', 'titi']
    assert shell_split('toto ""') == ['toto', '']
    assert shell_split('to"to titi "a" b') == ['to"to', 'titi', 'a', 'b']
    assert shell_split('"toto titi" toto ""') == ['toto titi', 'toto', '']
    assert shell_split('toto "titi') == ['toto', 'titi']

def test_argument_quoted():
    assert find_argument_quoted(4, 'toto titi tata') == 3
    assert find_argument_quoted(4, '"toto titi" tata') == 0
    assert find_argument_quoted(8, '"toto" "titi tata"') == 1
    assert find_argument_quoted(8, '"toto" "titi tata') == 1
    assert find_argument_quoted(3, '"toto" "titi tata') == 0
    assert find_argument_quoted(18, '"toto" "titi tata" ') == 2

def test_argument_unquoted():
    assert find_argument_unquoted(2, 'toto titi tata') == 0
    assert find_argument_unquoted(3, 'toto titi tata') == 0
    assert find_argument_unquoted(6, 'toto titi tata') == 1
    assert find_argument_unquoted(4, 'toto titi tata') == 3
    assert find_argument_unquoted(25, 'toto titi tata') == 3

def test_parse_str_to_secs():
    assert parse_str_to_secs("1d3m1h") == 90180
    assert parse_str_to_secs("1d3mfaiiiiil") == 0

def test_parse_secs_to_str():
    assert parse_secs_to_str(3601) == '1h1s'
    assert parse_secs_to_str(0) == '0s'

    with pytest.raises(TypeError):
        parse_secs_to_str('toto')

def test_unique_prefix_of__no_shared_prefix():
    assert unique_prefix_of("a", "b") == "a"
    assert unique_prefix_of("foo", "bar") == "f"
    assert unique_prefix_of("foo", "") == "f"

def test_unique_prefix_of__equal():
    assert unique_prefix_of("foo", "foo") == "foo"

def test_unique_prefix_of__a_prefix():
    assert unique_prefix_of("foo", "foobar") == "foo"

def test_unique_prefix_of__b_prefix():
    assert unique_prefix_of("foobar", "foo") == "foob"

def test_unique_prefix_of__normal_shared_prefix():
    assert unique_prefix_of("foobar", "foobaz") == "foobar"
    assert unique_prefix_of("fnord", "funky") == "fn"
    assert unique_prefix_of("asbestos", "aspergers") == "asb"



"""
A generic tab that displays a series of items in a scrollable, searchable,
sortable list.  It should be inherited, to actually provide methods that
insert items in the list, and that lets the user interact with them.
"""

import curses
import collections
import logging
from typing import Any, Callable, TYPE_CHECKING

from poezio import windows
from poezio.core.structs import Command
from poezio.tabs import DefaultHelpTab, Tab

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class ListTab(DefaultHelpTab):
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', name: str, help_message: str,
                 header_text: str, cols: list[tuple[str, int]]) -> None:
        """Parameters:
        name: The name of the tab
        help_message: The default help message displayed instead of the
        input
        header_text: The text displayed on the header line, at the top of
        the tab
        cols: a tuple of 2-tuples. e.g. (('column1_name', number),
                                         ('column2_name', number))
        """
        self.default_help_message = windows.HelpText(help_message)
        DefaultHelpTab.__init__(self, core)
        self.state = 'normal'
        self._error_message = ''
        self._name = name
        columns = collections.OrderedDict()
        for col, num in cols:
            columns[col] = num
        self.list_header = windows.ColumnHeaderWin(list(columns))
        self.listview = windows.ListWin(columns)
        self.info_header = windows.MucListInfoWin(header_text)
        self.key_func["KEY_DOWN"] = self.move_cursor_down
        self.key_func["KEY_UP"] = self.move_cursor_up
        self.key_func['^I'] = self.completion
        self.key_func["/"] = self.on_slash
        self.key_func['KEY_LEFT'] = self.list_header.sel_column_left
        self.key_func['KEY_RIGHT'] = self.list_header.sel_column_right
        self.key_func[' '] = self.sort_by
        self.register_command('close', self.close, shortdesc='Close this tab.')
        self.resize()
        self.update_keys()
        self.update_commands()

    def get_columns_sizes(self) -> dict[str, int]:
        """
        Must be implemented in subclasses. Must return a dict like this:
        {'column1_name': size1,
         'column2_name': size2}
        Where the size are calculated based on the size of the tab etc
        """
        raise NotImplementedError

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)
        if self.size.tab_degrade_y:
            display_info_win = False
        else:
            display_info_win = True

        self.info_header.refresh(window=self.listview)
        if display_info_win:
            self.info_win.refresh()
        self.refresh_tab_win()
        self.list_header.refresh()
        self.listview.refresh()
        self.input.refresh()

    def resize(self) -> None:
        if self.size.tab_degrade_y:
            info_win_height = 0
            tab_win_height = 0
        else:
            info_win_height = self.core.information_win_size
            tab_win_height = Tab.tab_win_height()

        self.info_header.resize(
            1, self.width, self.height - 2 - info_win_height - tab_win_height,
            0)
        column_size = self.get_columns_sizes()
        self.list_header.resize_columns(column_size)
        self.list_header.resize(1, self.width, 0, 0)
        self.listview.resize_columns(column_size)
        self.listview.resize(
            self.height - 3 - info_win_height - tab_win_height, self.width, 1,
            0)
        self.input.resize(1, self.width, self.height - 1, 0)

    def close(self, _: None = None) -> None:
        self.core.close_tab(self)

    def set_error(self, msg: str, code: str, body: str) -> None:
        """
        If there's an error (retrieving the values etc)
        """
        self._error_message = 'Error: %(code)s - %(msg)s: %(body)s' % {
            'msg': msg,
            'body': body,
            'code': code
        }
        self.info_header.message = self._error_message
        self.info_header.refresh()
        curses.doupdate()

    def sort_by(self) -> None:
        if self.list_header.get_order():
            self.listview.sort_by_column(
                col_name=self.list_header.get_sel_column(), asc=False)
            self.list_header.set_order(False)
            self.list_header.refresh()
        else:
            self.listview.sort_by_column(
                col_name=self.list_header.get_sel_column(), asc=True)
            self.list_header.set_order(True)
            self.list_header.refresh()
        self.core.doupdate()

    def on_info_win_size_changed(self) -> None:
        if self.core.information_win_size >= self.height - 3:
            return
        self.info_header.resize(
            1, self.width, self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), 0)
        self.listview.resize(
            self.height - 3 - self.core.information_win_size -
            Tab.tab_win_height(), self.width, 1, 0)

    def on_lose_focus(self) -> None:
        self.state = 'normal'

    def on_gain_focus(self) -> None:
        self.state = 'current'
        curses.curs_set(0)

    def on_scroll_up(self) -> bool:
        return self.listview.scroll_up()

    def on_scroll_down(self) -> bool:
        return self.listview.scroll_down()

    def move_cursor_up(self) -> None:
        self.listview.move_cursor_up()
        self.listview.refresh()
        self.core.doupdate()

    def move_cursor_down(self) -> None:
        self.listview.move_cursor_down()
        self.listview.refresh()
        self.core.doupdate()

    def matching_names(self) -> list[tuple[int, str]]:
        return [(2, self.name)]



#!/usr/bin/env python3
# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Define the variables (colors and some other stuff) that are
used when drawing the interface.

Colors are numbers from -1 to 7 (if only 8 colors are supported) or -1 to 255
if 256 colors are available.
If only 8 colors are available, all colors > 8 are converted using the
table_256_to_16 dict.

XHTML-IM colors are converted to -1 -> 255 colors if available, or directly to
-1 -> 8 if we are in 8-color-mode.

A pair_color is a background-foreground pair. All possible pairs are not created
at startup, because that would create 256*256 pairs, and almost all of them
would never be used.

A theme should define color tuples, like ``(200, -1)``, and when they are to
be used by poezio's interface, they will be created once, and kept in a list for
later usage.
A color tuple is of the form ``(foreground, background, optional)``
A color of -1 means the default color. So if you do not want to have
a background color, use ``(x, -1)``.
The optional third value of the tuple defines additional information. It
is a string and can contain one or more of these characters:

- ``b``: bold
- ``u``: underlined
- ``x``: blink

For example, ``(200, 208, 'bu')`` is bold, underlined and pink foreground on
orange background.

A theme file is a python file containing one object named 'theme', which is an
instance of a class (derived from the Theme class) defined in that same file.
For example, in pinkytheme.py:

.. code-block:: python

    import theming
    class PinkyTheme(theming.Theme):
        COLOR_NORMAL_TEXT = (200, -1)

    theme = PinkyTheme()

if the command '/theme pinkytheme' is issued, we import the pinkytheme.py file
and set the global variable 'theme' to pinkytheme.theme.

And in poezio's code we just use ``theme.COLOR_NORMAL_TEXT`` etc

Since a theme inherites from the Theme class (defined here), if a color is not defined in a
theme file, the color is the default one.

Some values in that class are a list of color tuple.
For example ``[(1, -1), (2, -1), (3, -1)]``
Such a list SHOULD contain at least one color tuple.
It is used for example to define color gradient, etc.
"""


import curses
import functools
import importlib.util
import logging
from datetime import datetime
from importlib import machinery
from os import path
from pathlib import Path
from typing import ClassVar

from poezio import colors, xdg
try:
    from poezio.config import config
except ImportError:
    if __name__ != "__main__":
        raise

finder = machinery.PathFinder()
log = logging.getLogger(__name__)


TextStyleNoAttr = tuple[int, int]
TextStyleAttr = tuple[int, int, str]
TextStyle = TextStyleNoAttr | TextStyleAttr


class Theme:
    """
    The theme class, from which all themes should inherit.
    All of the following values can be replaced in subclasses, in
    order to create a new theme.

    Do not edit this file if you want to change the theme to suit your
    needs. Create a new theme and share it if you think it can be useful
    for others.
    """

    @classmethod
    def color_role(cls, role: str) -> TextStyle:
        role_mapping = {
            'moderator': cls.COLOR_USER_MODERATOR,
            'participant': cls.COLOR_USER_PARTICIPANT,
            'visitor': cls.COLOR_USER_VISITOR,
            'none': cls.COLOR_USER_NONE,
            '': cls.COLOR_USER_NONE
        }
        return role_mapping.get(role, cls.COLOR_USER_NONE)

    @classmethod
    def char_affiliation(cls, affiliation: str) -> str:
        affiliation_mapping = {
            'owner': cls.CHAR_AFFILIATION_OWNER,
            'admin': cls.CHAR_AFFILIATION_ADMIN,
            'member': cls.CHAR_AFFILIATION_MEMBER,
            'none': cls.CHAR_AFFILIATION_NONE
        }
        return affiliation_mapping.get(affiliation, cls.CHAR_AFFILIATION_NONE)

    @classmethod
    def color_show(cls, show: str) -> TextStyle:
        show_mapping = {
            'xa': cls.COLOR_STATUS_XA,
            'none': cls.COLOR_STATUS_NONE,
            'dnd': cls.COLOR_STATUS_DND,
            'away': cls.COLOR_STATUS_AWAY,
            'chat': cls.COLOR_STATUS_CHAT,
            '': cls.COLOR_STATUS_ONLINE,
            'available': cls.COLOR_STATUS_ONLINE,
            'unavailable': cls.COLOR_STATUS_UNAVAILABLE,
        }
        return show_mapping.get(show, cls.COLOR_STATUS_NONE)

    @classmethod
    def char_subscription(cls, sub: str, keep: str = 'incomplete') -> str:
        sub_mapping = {
            'from': cls.CHAR_ROSTER_FROM,
            'both': cls.CHAR_ROSTER_BOTH,
            'none': cls.CHAR_ROSTER_NONE,
            'to': cls.CHAR_ROSTER_TO,
        }
        if keep == 'incomplete' and sub == 'both':
            return ''
        if keep in ('both', 'none', 'to', 'from'):
            return sub_mapping[sub] if sub == keep else ''
        return sub_mapping.get(sub, '')

    # Short date format (only show time)
    SHORT_TIME_FORMAT: ClassVar[str] = '%H:%M:%S'
    SHORT_TIME_FORMAT_LENGTH: ClassVar[int] = len(datetime.now().strftime(SHORT_TIME_FORMAT))

    # Long date format (show date and time)
    LONG_TIME_FORMAT: ClassVar[str] = '%Y-%m-%d %H:%M:%S'
    LONG_TIME_FORMAT_LENGTH: ClassVar[int] = len(datetime.now().strftime(LONG_TIME_FORMAT))

    # Message text color
    COLOR_NORMAL_TEXT: ClassVar[TextStyle] = (-1, -1)
    COLOR_INFORMATION_TEXT: ClassVar[TextStyle] = (5, -1)  # TODO
    COLOR_WARNING_TEXT: ClassVar[TextStyle] = (1, -1)
    COLOR_REACTIONS_TEXT: ClassVar[TextStyle] = COLOR_INFORMATION_TEXT

    # Color of the commands in the help message
    COLOR_HELP_COMMANDS: ClassVar[TextStyle] = (208, -1)

    # "reverse" is a special value, available only for this option. It just
    # takes the nick colors and reverses it. A theme can still specify a
    # fixed color if need be.
    COLOR_HIGHLIGHT_NICK: str | TextStyle = "reverse"

    # Color of the participant JID in a MUC
    COLOR_MUC_JID: ClassVar[TextStyle] = (4, -1)

    # User list color
    COLOR_USER_VISITOR: ClassVar[TextStyle] = (239, -1)
    COLOR_USER_PARTICIPANT: ClassVar[TextStyle] = (4, -1)
    COLOR_USER_NONE: ClassVar[TextStyle] = (0, -1)
    COLOR_USER_MODERATOR: ClassVar[TextStyle] = (1, -1)

    # nickname colors
    COLOR_REMOTE_USER: ClassVar[TextStyle] = (5, -1)

    # The character printed in color (COLOR_STATUS_*) before the nickname
    # in the user list
    CHAR_STATUS: ClassVar[str] = '|'

    # The characters used for the chatstates in the user list
    # in a MUC
    CHAR_CHATSTATE_ACTIVE: ClassVar[str] = 'A'
    CHAR_CHATSTATE_COMPOSING: ClassVar[str] = 'X'
    CHAR_CHATSTATE_PAUSED: ClassVar[str] = 'p'

    # These characters are used for the affiliation wherever needed, e.g., in
    # the user list in a MUC, or when displaying affiliation lists.
    CHAR_AFFILIATION_OWNER: ClassVar[str] = '~'
    CHAR_AFFILIATION_ADMIN: ClassVar[str] = '&'
    CHAR_AFFILIATION_MEMBER: ClassVar[str] = '+'
    CHAR_AFFILIATION_NONE: ClassVar[str] = '-'
    CHAR_AFFILIATION_OUTCAST: ClassVar[str] = '!'

    # XML Tab
    CHAR_XML_IN: ClassVar[str] = 'IN '
    CHAR_XML_OUT: ClassVar[str] = 'OUT'
    COLOR_XML_IN: ClassVar[TextStyle] = (1, -1)
    COLOR_XML_OUT: ClassVar[TextStyle] = (2, -1)

    # Color for the /me message
    COLOR_ME_MESSAGE: ClassVar[TextStyle] = (6, -1)

    # Color for the number of revisions of a message
    COLOR_REVISIONS_MESSAGE: ClassVar[TextStyle] = (3, -1, 'b')

    # Color for various important text. For example the "?" before JIDs in
    # the roster that require a user action.
    COLOR_IMPORTANT_TEXT: ClassVar[TextStyle] = (3, 5, 'b')

    # Separators
    COLOR_VERTICAL_SEPARATOR: ClassVar[TextStyle] = (4, -1)
    COLOR_NEW_TEXT_SEPARATOR: ClassVar[TextStyle] = (2, -1)
    COLOR_MORE_INDICATOR: ClassVar[TextStyle] = (6, 4)

    # Time
    CHAR_TIME_LEFT: ClassVar[str] = ''
    CHAR_TIME_RIGHT: ClassVar[str] = ''
    COLOR_TIME_STRING: ClassVar[TextStyle] = (-1, -1)

    # Tabs
    COLOR_TAB_NORMAL: ClassVar[TextStyle] = (7, 4)
    COLOR_TAB_NONEMPTY: ClassVar[TextStyle] = (7, 4)
    COLOR_TAB_SCROLLED: ClassVar[TextStyle] = (5, 4)
    COLOR_TAB_JOINED: ClassVar[TextStyle] = (82, 4)
    COLOR_TAB_CURRENT: ClassVar[TextStyle] = (7, 6)
    COLOR_TAB_COMPOSING: ClassVar[TextStyle] = (7, 5)
    COLOR_TAB_NEW_MESSAGE: ClassVar[TextStyle] = (7, 5)
    COLOR_TAB_HIGHLIGHT: ClassVar[TextStyle] = (7, 3)
    COLOR_TAB_PRIVATE: ClassVar[TextStyle] = (7, 2)
    COLOR_TAB_ATTENTION: ClassVar[TextStyle] = (7, 1)
    COLOR_TAB_DISCONNECTED: ClassVar[TextStyle] = (7, 8)

    # If autocolor_tab_names is set to true, the following modes are used to
    # distinguish tabs with normal and important messages.
    MODE_TAB_NORMAL: ClassVar[str] = ''
    MODE_TAB_IMPORTANT: ClassVar[str] = 'r'  # reverse video mode

    # This is the mode used for the tab name in the info bar of MUC and 1:1
    # chat tabs.
    MODE_TAB_NAME: ClassVar[str] = 'r'

    COLOR_VERTICAL_TAB_NORMAL = (4, -1)
    COLOR_VERTICAL_TAB_NONEMPTY = (4, -1)
    COLOR_VERTICAL_TAB_JOINED = (82, -1)
    COLOR_VERTICAL_TAB_SCROLLED = (66, -1)
    COLOR_VERTICAL_TAB_CURRENT = (7, 4)
    COLOR_VERTICAL_TAB_NEW_MESSAGE = (5, -1)
    COLOR_VERTICAL_TAB_COMPOSING = (5, -1)
    COLOR_VERTICAL_TAB_HIGHLIGHT = (3, -1)
    COLOR_VERTICAL_TAB_PRIVATE = (2, -1)
    COLOR_VERTICAL_TAB_ATTENTION = (1, -1)
    COLOR_VERTICAL_TAB_DISCONNECTED = (8, -1)

    # yapf: disable

    # Nickname colors
    # A list of colors randomly attributed to nicks in MUCs
    # Setting more colors makes it harder to have two nicks with the same color,
    # avoiding confusions.
    LIST_COLOR_NICKNAMES: ClassVar[list[TextStyleNoAttr]] = [
            (1, -1), (2, -1), (3, -1), (4, -1), (5, -1), (6, -1), (9, -1),
            (10, -1), (11, -1), (12, -1), (13, -1), (14, -1), (19, -1),
            (20, -1), (21, -1), (22, -1), (23, -1), (24, -1), (25, -1),
            (26, -1), (27, -1), (28, -1), (29, -1), (30, -1), (31, -1),
            (32, -1), (33, -1), (34, -1), (35, -1), (36, -1), (37, -1),
            (38, -1), (39, -1), (40, -1), (41, -1), (42, -1), (43, -1),
            (44, -1), (45, -1), (46, -1), (47, -1), (48, -1), (49, -1),
            (50, -1), (51, -1), (54, -1), (55, -1), (56, -1), (57, -1),
            (58, -1), (60, -1), (61, -1), (62, -1), (63, -1), (64, -1),
            (65, -1), (66, -1), (67, -1), (68, -1), (69, -1), (70, -1),
            (71, -1), (72, -1), (73, -1), (74, -1), (75, -1), (76, -1),
            (77, -1), (78, -1), (79, -1), (80, -1), (81, -1), (82, -1),
            (83, -1), (84, -1), (85, -1), (86, -1), (87, -1), (88, -1),
            (89, -1), (90, -1), (91, -1), (92, -1), (93, -1), (94, -1),
            (95, -1), (96, -1), (97, -1), (98, -1), (99, -1), (100, -1),
            (101, -1), (103, -1), (104, -1), (105, -1), (106, -1), (107, -1),
            (108, -1), (109, -1), (110, -1), (111, -1), (112, -1), (113, -1),
            (114, -1), (115, -1), (116, -1), (117, -1), (118, -1), (119, -1),
            (120, -1), (121, -1), (122, -1), (123, -1), (124, -1), (125, -1),
            (126, -1), (127, -1), (128, -1), (129, -1), (130, -1), (131, -1),
            (132, -1), (133, -1), (134, -1), (135, -1), (136, -1), (137, -1),
            (138, -1), (139, -1), (140, -1), (141, -1), (142, -1), (143, -1),
            (144, -1), (145, -1), (146, -1), (147, -1), (148, -1), (149, -1),
            (150, -1), (151, -1), (152, -1), (153, -1), (154, -1), (155, -1),
            (156, -1), (157, -1), (158, -1), (159, -1), (160, -1), (161, -1),
            (162, -1), (163, -1), (164, -1), (165, -1), (166, -1), (167, -1),
            (168, -1), (169, -1), (170, -1), (171, -1), (172, -1), (173, -1),
            (174, -1), (175, -1), (176, -1), (177, -1), (178, -1), (179, -1),
            (180, -1), (181, -1), (182, -1), (183, -1), (184, -1), (185, -1),
            (186, -1), (187, -1), (188, -1), (189, -1), (190, -1), (191, -1),
            (192, -1), (193, -1), (196, -1), (197, -1), (198, -1), (199, -1),
            (200, -1), (201, -1), (202, -1), (203, -1), (204, -1), (205, -1),
            (206, -1), (207, -1), (208, -1), (209, -1), (210, -1), (211, -1),
            (212, -1), (213, -1), (214, -1), (215, -1), (216, -1), (217, -1),
            (218, -1), (219, -1), (220, -1), (221, -1), (222, -1), (223, -1),
            (224, -1), (225, -1), (226, -1), (227, -1)]
    # XEP-0392 consistent color generation palette placeholder
    # it’s generated on first use when accessing the ccg_palette property
    CCG_PALETTE: dict[float, int] | None = None
    CCG_Y: float = 0.5**0.45

    # yapf: enable

    # This is your own nickname
    COLOR_OWN_NICK: ClassVar[TextStyle] = (254, -1)

    COLOR_LOG_MSG: ClassVar[TextStyle] = (5, -1)
    # This is for in-tab error messages
    COLOR_ERROR_MSG: ClassVar[TextStyle] = (9, 7, 'b')
    # Status color
    COLOR_STATUS_XA: ClassVar[TextStyle] = (16, 90)
    COLOR_STATUS_NONE: ClassVar[TextStyle] = (16, 4)
    COLOR_STATUS_DND: ClassVar[TextStyle] = (16, 1)
    COLOR_STATUS_AWAY: ClassVar[TextStyle] = (16, 3)
    COLOR_STATUS_CHAT: ClassVar[TextStyle] = (16, 2)
    COLOR_STATUS_UNAVAILABLE: ClassVar[TextStyle] = (-1, 247)
    COLOR_STATUS_ONLINE: ClassVar[TextStyle] = (16, 4)

    # Bars
    COLOR_WARNING_PROMPT: ClassVar[TextStyle] = (16, 1, 'b')
    COLOR_INFORMATION_BAR: ClassVar[TextStyle] = (7, 4)
    COLOR_TOPIC_BAR: ClassVar[TextStyle] = (7, 4)
    COLOR_SCROLLABLE_NUMBER: ClassVar[TextStyle] = (220, 4, 'b')
    COLOR_SELECTED_ROW: ClassVar[TextStyle] = (-1, 33)
    COLOR_PRIVATE_NAME: ClassVar[TextStyle] = (-1, 4)
    COLOR_CONVERSATION_NAME: ClassVar[TextStyle] = (2, 4)
    COLOR_CONVERSATION_RESOURCE: ClassVar[TextStyle] = (121, 4)
    COLOR_GROUPCHAT_NAME: ClassVar[TextStyle] = (7, 4)
    COLOR_COLUMN_HEADER: ClassVar[TextStyle] = (36, 4)
    COLOR_COLUMN_HEADER_SEL: ClassVar[TextStyle] = (4, 36)

    # Strings for special messages (like join, quit, nick change, etc)
    CHAR_BEFORE_NICK_ME: ClassVar[str] = '* '
    CHAR_AFTER_NICK_ME: ClassVar[str] = ' '
    CHAR_AFTER_NICK: ClassVar[str] = '> '
    CHAR_JOIN: ClassVar[str] = '--->'
    CHAR_QUIT: ClassVar[str] = '<---'
    CHAR_KICK: ClassVar[str] = '-!-'
    CHAR_NEW_TEXT_SEPARATOR: ClassVar[str] = '- '
    CHAR_OK: ClassVar[str] = '✔'
    CHAR_ERROR: ClassVar[str] = '✖'
    CHAR_EMPTY: ClassVar[str] = ' '
    CHAR_ACK_RECEIVED: ClassVar[str] = CHAR_OK
    CHAR_NACK: ClassVar[str] = CHAR_ERROR
    CHAR_COLUMN_ASC: ClassVar[str] = ' ▲'
    CHAR_COLUMN_DESC: ClassVar[str] = ' ▼'
    CHAR_ROSTER_ERROR: ClassVar[str] = CHAR_ERROR
    CHAR_ROSTER_TUNE: ClassVar[str] = '♪'
    CHAR_ROSTER_ASKED: ClassVar[str] = '?'
    CHAR_ROSTER_ACTIVITY: ClassVar[str] = 'A'
    CHAR_ROSTER_MOOD: ClassVar[str] = 'M'
    CHAR_ROSTER_GAMING: ClassVar[str] = 'G'
    CHAR_ROSTER_FROM: ClassVar[str] = '←'
    CHAR_ROSTER_BOTH: ClassVar[str] = '↔'
    CHAR_ROSTER_TO: ClassVar[str] = '→'
    CHAR_ROSTER_NONE: ClassVar[str] = '⇹'

    COLOR_CHAR_ACK: ClassVar[TextStyle] = (2, -1)
    COLOR_CHAR_NACK: ClassVar[TextStyle] = (1, -1)

    COLOR_ROSTER_GAMING: ClassVar[TextStyle] = (6, -1)
    COLOR_ROSTER_MOOD: ClassVar[TextStyle] = (2, -1)
    COLOR_ROSTER_ACTIVITY: ClassVar[TextStyle] = (3, -1)
    COLOR_ROSTER_TUNE: ClassVar[TextStyle] = (6, -1)
    COLOR_ROSTER_ERROR: ClassVar[TextStyle] = (1, -1)
    COLOR_ROSTER_SUBSCRIPTION: ClassVar[TextStyle] = (-1, -1)

    COLOR_JOIN_CHAR: ClassVar[TextStyle] = (4, -1)
    COLOR_QUIT_CHAR: ClassVar[TextStyle] = (1, -1)
    COLOR_KICK_CHAR: ClassVar[TextStyle] = (1, -1)

    # Vertical tab list color
    COLOR_VERTICAL_TAB_NUMBER: ClassVar[TextStyle] = (34, -1)

    # Info messages color (the part before the ">")
    INFO_COLORS: ClassVar[dict[str, TextStyle]] = {
        'info': (5, -1),
        'error': (9, 7, 'b'),
        'warning': (1, -1),
        'roster': (2, -1),
        'help': (10, -1),
        'headline': (11, -1, 'b'),
        'tune': (6, -1),
        'gaming': (6, -1),
        'mood': (2, -1),
        'activity': (3, -1),
        'default': (7, -1),
    }

    @property
    def ccg_palette(self) -> dict[float, int] | None:
        prepare_ccolor_palette(self)
        return self.CCG_PALETTE


# This is the default theme object, used if no theme is defined in the conf
theme = Theme()

# a dict "color tuple -> color_pair"
# Each time we use a color tuple, we check if it has already been used.
# If not we create a new color_pair and keep it in that dict, to use it
# the next time.
curses_colors_dict: dict[TextStyle, int] = {}

# yapf: disable

table_256_to_16 = [
    0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10, 11, 12, 13, 14, 15,
    0,  4,  4,  4,  12, 12, 2,  6,  4,  4,  12, 12, 2,  2,  6,  4,
    12, 12, 2,  2,  2,  6,  12, 12, 10, 10, 10, 10, 14, 12, 10, 10,
    10, 10, 10, 14, 1,  5,  4,  4,  12, 12, 3,  8,  4,  4,  12, 12,
    2,  2,  6,  4,  12, 12, 2,  2,  2,  6,  12, 12, 10, 10, 10, 10,
    14, 12, 10, 10, 10, 10, 10, 14, 1,  1,  5,  4,  12, 12, 1,  1,
    5,  4,  12, 12, 3,  3,  8,  4,  12, 12, 2,  2,  2,  6,  12, 12,
    10, 10, 10, 10, 14, 12, 10, 10, 10, 10, 10, 14, 1,  1,  1,  5,
    12, 12, 1,  1,  1,  5,  12, 12, 1,  1,  1,  5,  12, 12, 3,  3,
    3,  7,  12, 12, 10, 10, 10, 10, 14, 12, 10, 10, 10, 10, 10, 14,
    9,  9,  9,  9,  13, 12, 9,  9,  9,  9,  13, 12, 9,  9,  9,  9,
    13, 12, 9,  9,  9,  9,  13, 12, 11, 11, 11, 11, 7,  12, 10, 10,
    10, 10, 10, 14, 9,  9,  9,  9,  9,  13, 9,  9,  9,  9,  9,  13,
    9,  9,  9,  9,  9,  13, 9,  9,  9,  9,  9,  13, 9,  9,  9,  9,
    9,  13, 11, 11, 11, 11, 11, 15, 0,  0,  0,  0,  0,  0,  8,  8,
    8,  8,  8,  8,  7,  7,  7,  7,  7,  7,  15, 15, 15, 15, 15, 15
]
# yapf: enable

load_path: list[str] = []


def color_256_to_16(color: int) -> int:
    if color == -1:
        return color
    return table_256_to_16[color]


def dump_tuple(tup: TextStyle) -> str:
    """
    Dump a tuple to a string of fg,bg,attr (optional)
    """
    return ','.join(str(i) for i in tup)


def read_tuple(_str: str) -> tuple[TextStyleNoAttr, str]:
    """
    Read a tuple dumped with dump_tumple
    """
    attrs = _str.split(',')
    char = attrs[2] if len(attrs) > 2 else '\0'
    return (int(attrs[0]), int(attrs[1])), char


@functools.lru_cache(maxsize=128)
def to_curses_attr(color_tuple: TextStyle) -> int:
    """
    Takes a color tuple (as defined at the top of this file) and
    returns a valid curses attr that can be passed directly to attron() or attroff()
    """
    # extract the color from that tuple
    colors: TextStyle
    if len(color_tuple) == 3:
        colors = (color_tuple[0], color_tuple[1])
    else:
        colors = color_tuple

    bold = False
    if curses.COLORS < 256:
        # We are not in a term supporting 256 colors, so we convert
        # colors to numbers between -1 and 8
        colors = (color_256_to_16(colors[0]), color_256_to_16(colors[1]))
        if colors[0] >= 8:
            colors = (colors[0] - 8, colors[1])
            bold = True
        if colors[1] >= 8:
            colors = (colors[0], colors[1] - 8)

    # check if we already used these colors
    try:
        pair = curses_colors_dict[colors]
    except KeyError:
        pair = len(curses_colors_dict) + 1
        curses.init_pair(pair, colors[0], colors[1])
        curses_colors_dict[colors] = pair
    curses_pair = curses.color_pair(pair)
    if len(color_tuple) == 3:
        _, _, additional_val = color_tuple
        if 'b' in additional_val or bold is True:
            curses_pair = curses_pair | curses.A_BOLD
        if 'u' in additional_val:
            curses_pair = curses_pair | curses.A_UNDERLINE
        if 'i' in additional_val:
            curses_pair = curses_pair | (curses.A_ITALIC if hasattr(
                curses, 'A_ITALIC') else curses.A_REVERSE)
        if 'a' in additional_val:
            curses_pair = curses_pair | curses.A_BLINK
        if 'r' in additional_val:
            curses_pair = curses_pair | curses.A_REVERSE
    return curses_pair


def get_theme() -> Theme:
    """
    Returns the current theme
    """
    return theme


def update_themes_dir(option: str | None = None,
                      value: str | None = None) -> None:
    global load_path
    load_path = []

    # import from the git sources
    default_dir = path.join(
        path.dirname(path.dirname(__file__)), 'data/themes')
    if path.exists(default_dir):
        load_path.append(default_dir)

    # import from the user-defined prefs
    themes_dir_str = config.getstr('themes_dir')
    themes_dir = Path(themes_dir_str).expanduser(
    ) if themes_dir_str else xdg.DATA_HOME / 'themes'
    try:
        themes_dir.mkdir(parents=True, exist_ok=True)
    except OSError:
        log.exception('Unable to create the themes dir (%s):', themes_dir)
    else:
        load_path.append(str(themes_dir))

    # system-wide import
    try:  # noqa: TRY101
        poezio_themes = importlib.import_module('poezio_themes')
    except ImportError:
        pass
    else:
        if poezio_themes and poezio_themes.__path__:
            load_path.append(list(poezio_themes.__path__)[0])

    log.debug('Theme load path: %s', load_path)


def prepare_ccolor_palette(theme: Theme) -> None:
    """
    Prepare the Consistent Color Generation (XEP-0392) palette for a theme.
    """
    if theme.CCG_PALETTE is not None:
        return None

    if any(bg != -1 for fg, bg in theme.LIST_COLOR_NICKNAMES):
        # explicitly disable CCG, can’t handle dynamic background colors
        theme.CCG_PALETTE = {}
        return None

    theme.CCG_PALETTE = colors.generate_ccg_palette(
        [
            fg for fg, _ in theme.LIST_COLOR_NICKNAMES
            # exclude grayscale
            if fg < 232
        ],
        theme.CCG_Y,
    )
    return None


def reload_theme() -> str | None:
    theme_name = config.getstr('theme')
    global theme
    if theme_name == 'default' or not theme_name.strip():
        theme = Theme()
        return None
    new_theme = None
    exc = None
    try:
        spec = finder.find_spec(theme_name, path=load_path)
        if spec and spec.loader:
            new_theme = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(new_theme)
        else:
            return 'Failed to load the theme %s' % theme_name
        new_theme = spec.loader.load_module(theme_name)
    except Exception as e:
        log.exception('Failed to load the theme %s', theme_name)
        exc = e

    if not new_theme:
        return 'Failed to load theme: %s' % exc

    if hasattr(new_theme, 'theme'):
        theme = new_theme.theme
        prepare_ccolor_palette(theme)
        return None
    return 'No theme present in the theme file'


if __name__ == '__main__':
    # Display some nice text with nice colors
    s = curses.initscr()
    curses.start_color()
    curses.use_default_colors()
    s.addstr('%s colors detected\n\n' % curses.COLORS, to_curses_attr((3, -1)))
    for i in range(curses.COLORS):
        s.addstr('%s ' % i, to_curses_attr((i, -1)))
    s.addstr('\n')
    s.refresh()
    try:
        s.getkey()
    except KeyboardInterrupt:
        pass
    finally:
        curses.endwin()
        print()



"""
This plugin allows you to quote messages easily.

Usage
-----

.. glossary::

    /quote
        **Usage:** ``/quote <message>``

        The message must exist. You can use autocompletion to get the message
        you want to quote easily.

        Example:

        .. code-block:: none

            /quote "Pouet"

        If the message "Pouet" exists, it will be put in the input. If not you
        will get a warning.

Options
-------

.. glossary::
    :sorted:

    before_quote

        **Default value:** ``[empty]``

        Text to insert before the quote. ``%(nick)s`` and ``%(time)s`` can
        be used to insert the nick of the user who sent the message or the
        time of the message.

    after_quote

        **Default value:** ``[empty]``

        Text to insert after the quote. ``%(nick)s`` and ``%(time)s`` can
        be used to insert the nick of the user who sent the message or the
        time of the message.
"""

from poezio.core.structs import Completion
from poezio.ui.types import Message, BaseMessage
from poezio.plugin import BasePlugin
from poezio.xhtml import clean_text
from poezio.theming import get_theme
from poezio.windows.inputs import Input
from poezio import common
from poezio import tabs

import logging
log = logging.getLogger(__name__)


class Plugin(BasePlugin):
    def init(self) -> None:
        for _class in (tabs.MucTab, tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.PrivateTab):
            self.api.add_tab_command(
                _class,
                'quote',
                self.command_quote,
                usage='<message>',
                help='Quote the message you typed if it exists.',
                short='Quote a message.',
                completion=self.completion_quote)

    def command_quote(self, str_args: str) -> None:
        args = common.shell_split(str_args)
        if len(args) == 1:
            message_time = args[-1]
        else:
            self.api.run_command('/help quote')
            return
        message = self.find_message(message_time)
        if message:
            str_time = message.time.strftime(get_theme().SHORT_TIME_FORMAT)
            before = self.config.get('before_quote', '') % {
                'nick': message.nickname or '',
                'time': str_time,
            }
            after = self.config.get('after_quote', '') % {
                'nick': message.nickname or '',
                'time': str_time,
            }
            self.core.insert_input_text(
                '%(before)s%(quote)s%(after)s' % {
                    'before': before.replace('\\n', '\n').replace('[SP]', ' '),
                    'quote': clean_text(message.txt),
                    'after': after.replace('\\n', '\n').replace('[SP]', ' ')
                })
        else:
            self.api.information('No message found', 'Warning')

    def find_message(self, txt: str) -> Message | None:
        messages = self.api.get_conversation_messages()
        if not messages:
            return None
        for message in messages[::-1]:
            if isinstance(message, Message) and clean_text(message.txt) == txt:
                return message
        return None

    def completion_quote(self, the_input: Input) -> Completion | None:
        messages = self.api.get_conversation_messages()
        if not messages:
            return None
        text = the_input.get_text()
        args = common.shell_split(text)
        if not text.endswith(' '):
            input_message = args[-1]

            def message_match(msg: BaseMessage) -> bool:
                return input_message.lower() in clean_text(msg.txt).lower()

            messages = list(filter(message_match, messages))
        elif len(args) > 1:
            return None
        return Completion(
            the_input.auto_completion,
            [clean_text(msg.txt) for msg in messages[::-1] if isinstance(msg, Message)],
            ''
        )



"""
Tests for the User class
"""

import pytest
from datetime import datetime
from slixmpp import JID
from poezio.user import User


@pytest.fixture
def user1():
    return User(
        nick='nick1',
        affiliation='member',
        show='xa',
        status='My Status!',
        role='moderator',
        jid=JID('foo@muc/nick1'),
        color='red',
    )


def test_new_user(user1):
    assert user1.last_talked == datetime(1, 1, 1)
    assert user1.jid == JID('foo@muc/nick1')
    assert user1.chatstate is None
    assert user1.affiliation == 'member'
    assert user1.show == 'xa'
    assert user1.status == 'My Status!'
    assert user1.role == 'moderator'
    assert user1.nick == 'nick1'
    assert user1.color == (196, -1)
    assert str(user1) == '>nick1<'


def test_change_nick(user1):
    user1.change_nick('nick2')
    assert user1.nick == 'nick2'


def test_change_color(user1):
    user1.change_color('blue')
    assert user1.color == (21, -1)



class Tab(object): pass
class ChatTab(Tab): pass
class PrivateTab(ChatTab): pass
class MucTab(ChatTab): pass
class ConversationTab(ChatTab): pass
class RosterInfoTab(Tab): pass
class XMLTab(Tab): pass
class DynamicConversationTab(Tab): pass
class StaticConversationTab(Tab): pass
class GapTab(Tab): pass




"""
Module for the ConversationTabs

A ConversationTab is a direct chat between two JIDs, outside of a room.

There are two different instances of a ConversationTab:
- A DynamicConversationTab that used to implement XEP-0296 (best
    practices for resource locking), which now stays on the bare JID at
    any time. This is the default.
- A StaticConversationTab that will stay focused on one resource all
    the time.

"""
import curses
import logging
from datetime import datetime
from typing import Any, Callable, Optional, TYPE_CHECKING

from slixmpp import JID, InvalidJID, Message as SMessage

from poezio.tabs.basetabs import OneToOneTab, Tab

from poezio import common
from poezio import windows
from poezio import xhtml
from poezio.config import config, get_image_cache
from poezio.core.structs import Command
from poezio.decorators import refresh_wrapper
from poezio.log_loader import MAMFiller
from poezio.logger import logger
from poezio.roster import roster
from poezio.theming import get_theme, dump_tuple
from poezio.decorators import command_args_parser
from poezio.ui.types import InfoMessage, Message
from poezio.text_buffer import CorrectionError

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from poezio.core.core import Core


class ConversationTab(OneToOneTab):
    """
    The tab containing a normal conversation (not from a MUC)
    Must not be instantiated, use Static or Dynamic version only.
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}
    additional_information: dict[str, Callable[[str], str]] = {}
    message_type = 'chat'
    nick: str | None

    def __init__(self, core: 'Core', jid: JID, initial: SMessage | None = None) -> None:
        OneToOneTab.__init__(self, core, jid, initial=initial)
        self.nick = None
        self.nick_sent = False
        self.state = 'normal'
        self.upper_bar = windows.ConversationStatusMessageWin()
        # keys
        self.key_func['^I'] = self.completion
        # commands
        self.register_command(
            'version',
            self.command_version,
            desc=(
                'Get the software version of the current interlocutor '
                '(usually their XMPP client and Operating System).'
            ),
            shortdesc='Get the software version of the user.')
        self.register_command(
            'info',
            self.command_info,
            shortdesc='Get the status of the contact.')
        self.register_command(
            'last_activity',
            self.command_last_activity,
            usage='[jid]',
            desc='Get the last activity of the given or the current contact.',
            shortdesc='Get the activity.',
            completion=self.core.completion.last_activity)
        self.register_command(
            'invite',
            self.core.command.impromptu,
            desc='Invite people into an impromptu room.',
            shortdesc='Invite other users to the discussion',
            completion=self.core.completion.impromptu)
        self.update_commands()
        self.update_keys()

    async def init_logs(self, initial: Optional[SMessage] = None) -> None:
        use_log = config.get_by_tabname('use_log', self.jid)
        mam_sync = config.get_by_tabname('mam_sync', self.jid)
        if use_log and mam_sync:
            limit = config.get_by_tabname('mam_sync_limit', self.jid)
            mam_filler = MAMFiller(logger, self, limit)
            self.mam_filler = mam_filler

            if initial is not None:
                # If there is an initial message, throw it back into the
                # text buffer if it cannot be fetched from mam
                await mam_filler.done.wait()
                if mam_filler.result == 0 and initial:
                    await self.handle_message(initial)
                initial = None
        await super().init_logs(initial=initial)

    @property
    def general_jid(self) -> JID:
        return JID(self.jid.bare)

    def get_info_header(self) -> windows.ConversationInfoWin:
        raise NotImplementedError

    @staticmethod
    @refresh_wrapper.always
    def add_information_element(plugin_name: str, callback: Callable[[str], str]) -> None:
        """
        Lets a plugin add its own information to the ConversationInfoWin
        """
        ConversationTab.additional_information[plugin_name] = callback

    @staticmethod
    @refresh_wrapper.always
    def remove_information_element(plugin_name: str) -> None:
        del ConversationTab.additional_information[plugin_name]

    def completion(self) -> None:
        self.complete_commands(self.input)

    async def handle_message(self, msg: SMessage, display: bool = True) -> None:
        """Handle a received message.

        The message can come from us (carbon copy).
        """

        # Prevent messages coming from our own devices (1:1) to be reflected
        if msg['to'].bare == self.core.xmpp.boundjid.bare and \
           msg['from'].bare == self.core.xmpp.boundjid.bare:
            _, index = self._text_buffer._find_message(msg['id'])
            if index != -1:
                return

        use_xhtml = config.get_by_tabname(
            'enable_xhtml_im',
            msg['from'].bare
        )
        tmp_dir = get_image_cache()

        # normal message, we are the recipient
        if msg['to'].bare == self.core.xmpp.boundjid.bare:
            conv_jid = msg['from']
            jid = conv_jid
            color = get_theme().COLOR_REMOTE_USER
            self.last_remote_message = datetime.now()
            remote_nick = self.get_nick()
        # we wrote the message (happens with carbons)
        elif msg['from'].bare == self.core.xmpp.boundjid.bare:
            conv_jid = msg['to']
            jid = self.core.xmpp.boundjid
            color = get_theme().COLOR_OWN_NICK
            remote_nick = self.core.own_nick
        # we are not part of that message, drop it
        else:
            return

        await self.core.events.trigger_async('conversation_msg', msg, self)

        if not msg['body']:
            return
        body = xhtml.get_body_from_message_stanza(
            msg, use_xhtml=use_xhtml, extract_images_to=tmp_dir)
        delayed, date = common.find_delayed_tag(msg)

        replaced = False
        if msg.get_plugin('replace', check=True):
            replaced_id = msg['replace']['id']
            if replaced_id and config.get_by_tabname('group_corrections',
                                                     conv_jid.bare):
                try:
                    replaced = self.modify_message(
                        body,
                        replaced_id,
                        msg['id'],
                        time=date,
                        jid=jid,
                        nickname=remote_nick)
                except CorrectionError:
                    log.debug('Unable to correct the message: %s', msg)
        if not replaced:
            message = Message(
                txt=body,
                time=date,
                nickname=remote_nick,
                nick_color=color,
                history=delayed,
                identifier=msg['id'],
                jid=jid,
            )
            if display:
                self.add_message(message)
            else:
                self.log_message(message)

    @refresh_wrapper.always
    @command_args_parser.raw
    async def command_say(self, line: str, attention: bool = False, correct: bool = False) -> None:
        await self._initial_log.wait()
        msg: SMessage = self.core.xmpp.make_message(
            mto=self.get_dest_jid(),
            mfrom=self.core.xmpp.boundjid
        )
        msg['type'] = 'chat'
        msg['body'] = line
        if not self.nick_sent:
            msg['nick'] = self.core.own_nick
            self.nick_sent = True
        # trigger the event BEFORE looking for colors.
        # and before displaying the message in the window
        # This lets a plugin insert \x19xxx} colors, that will
        # be converted in xhtml.
        self.core.events.trigger('conversation_say', msg, self)
        if not msg['body']:
            return
        if correct or msg['replace']['id']:
            msg['replace']['id'] = self.last_sent_message['id']  # type: ignore
        else:
            del msg['replace']
        if msg['body'].find('\x19') != -1:
            msg.enable('html')
            msg['html']['body'] = xhtml.poezio_colors_to_html(msg['body'])
            msg['body'] = xhtml.clean_text(msg['body'])
        if config.get_by_tabname('send_chat_states', self.general_jid):
            if self.inactive:
                self.send_chat_state('inactive', always_send=True)
            else:
                msg['chat_state'] = 'active'
        if attention:
            msg['attention'] = True
        self.core.events.trigger('conversation_say_after', msg, self)
        if not msg['body']:
            return
        self.set_last_sent_message(msg, correct=correct)
        msg._add_receipt = True  # type: ignore
        msg.send()
        await self.core.handler.on_normal_message(msg)
        # Our receipts slixmpp hack
        self.cancel_paused_delay()

    @command_args_parser.quoted(0, 1)
    async def command_last_activity(self, args: list[str]) -> None:
        """
        /last_activity [jid]
        """
        if args and args[0]:
            return await self.core.command.last_activity(args[0])

        iq = await self.core.xmpp.plugin['xep_0012'].get_last_activity(
            self.get_dest_jid(),
        )
        if iq['type'] != 'result':
            if iq['error']['type'] == 'auth':
                self.information(
                    'You are not allowed to see the activity of this contact.',
                    'Error',
                )
            else:
                self.information(
                    'Error retrieving the activity',
                    'Error',
                )
            return
        seconds = iq['last_activity']['seconds']
        status = iq['last_activity']['status']
        from_ = iq['from']
        msg = '\x19%s}The last activity of %s was %s ago%s'
        user = ''
        try:
            user = JID(from_).user
        except InvalidJID:
            pass

        if not user:
            msg = '\x19%s}The uptime of %s is %s.' % (
                dump_tuple(get_theme().COLOR_INFORMATION_TEXT), from_,
                common.parse_secs_to_str(seconds))
        else:
            msg = '\x19%s}The last activity of %s was %s ago%s' % (
                dump_tuple(get_theme().COLOR_INFORMATION_TEXT),
                from_,
                common.parse_secs_to_str(seconds),
                (' and their last status was %s' % status)
                if status else '',
            )
        self.add_message(InfoMessage(msg))
        self.core.refresh_window()

    @refresh_wrapper.conditional
    @command_args_parser.ignored
    def command_info(self) -> None:
        contact = roster[self.get_dest_jid()]
        try:
            jid = JID(self.get_dest_jid())
        except InvalidJID:
            jid = JID('')
        if contact:
            if jid.resource:
                resource = contact[jid.full]
            else:
                resource = contact.get_highest_priority_resource()
        else:
            resource = None
        if resource:
            status = (f', Status: {resource.status}') if resource.status else ''
            show = f"Show: {resource.presence or 'available'}"
            self.add_message(InfoMessage(f'{show}{status}'))
            return
        self.add_message(
            InfoMessage("No information available"),
        )
        return

    @command_args_parser.quoted(0, 1)
    async def command_version(self, args: list[str]) -> None:
        """
        /version [jid]
        """
        if args:
            await self.core.command.version(args[0])
            return
        jid = self.jid
        if not jid.resource:
            if jid in roster:
                resource = roster.contacts[jid].get_highest_priority_resource()
                jid = JID(resource.jid) if resource else JID(jid)
        iq = await self.core.xmpp.plugin['xep_0092'].get_version(jid)
        self.core.handler.on_version_result(iq)

    def resize(self) -> None:
        self.need_resize = False
        if self.size.tab_degrade_y:
            display_bar = False
            info_win_height = 0
            tab_win_height = 0
            bar_height = 0
        else:
            display_bar = True
            info_win_height = self.core.information_win_size
            tab_win_height = Tab.tab_win_height()
            bar_height = 1

        self.text_win.resize(
            self.height - 2 - bar_height - info_win_height - tab_win_height,
            self.width, bar_height, 0, self._text_buffer,
            force=self.ui_config_changed
        )
        self.ui_config_changed = False
        if display_bar:
            self.upper_bar.resize(1, self.width, 0, 0)
        self.get_info_header().resize(
            1, self.width, self.height - 2 - info_win_height - tab_win_height,
            0)
        self.input.resize(1, self.width, self.height - 1, 0)

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)
        display_bar = display_info_win = not self.size.tab_degrade_y

        self.text_win.refresh()

        if display_bar:
            self.upper_bar.refresh(self.get_dest_jid(),
                                   roster.contacts.get(self.get_dest_jid()))
        self.get_info_header().refresh(
            self.get_dest_jid(), roster[self.get_dest_jid()], self.text_win,
            self.chatstate, ConversationTab.additional_information)

        if display_info_win:
            self.info_win.refresh()
        self.refresh_tab_win()
        self.input.refresh()

    def refresh_info_header(self) -> None:
        self.get_info_header().refresh(
            self.get_dest_jid(), roster[self.get_dest_jid()], self.text_win,
            self.chatstate, ConversationTab.additional_information)
        self.input.refresh()

    def get_nick(self) -> str:
        contact = roster.contacts.get(self.jid.bare)
        if contact:
            return contact.name or self.jid.user
        else:
            if self.nick:
                return self.nick
            return self.jid.user or self.jid.domain

    def on_input(self, key: str, raw: bool) -> None:
        if not raw and key in self.key_func:
            self.key_func[key]()
            return
        self.input.do_command(key, raw=raw)
        empty_after = self.input.get_text() == '' or (
            self.input.get_text().startswith('/')
            and not self.input.get_text().startswith('//'))
        self.send_composing_chat_state(empty_after)

    def on_lose_focus(self) -> None:
        contact = roster[self.get_dest_jid()]
        try:
            jid = JID(self.get_dest_jid())
        except InvalidJID:
            jid = JID('')
        if contact:
            if jid.resource:
                resource = contact[jid.full]
            else:
                resource = contact.get_highest_priority_resource()
        else:
            resource = None
        if self.input.text:
            self.state = 'nonempty'
        else:
            self.state = 'normal'
        self.text_win.remove_line_separator()
        self.text_win.add_line_separator(self._text_buffer)
        if config.get_by_tabname('send_chat_states', self.general_jid):
            if resource:
                self.send_chat_state('inactive')
        self.check_scrolled()

    def on_gain_focus(self) -> None:
        contact = roster[self.get_dest_jid()]
        try:
            jid = JID(self.get_dest_jid())
        except InvalidJID:
            jid = JID('')
        if contact:
            if jid.resource:
                resource = contact[jid.full]
            else:
                resource = contact.get_highest_priority_resource()
        else:
            resource = None

        self.state = 'current'
        curses.curs_set(1)
        if (config.get_by_tabname('send_chat_states', self.general_jid)
                and (not self.input.get_text()
                     or not self.input.get_text().startswith('//'))):
            if resource:
                self.send_chat_state('active')

    def on_info_win_size_changed(self) -> None:
        if self.core.information_win_size >= self.height - 3:
            return
        self.text_win.resize(
            self.height - 3 - self.core.information_win_size -
            Tab.tab_win_height(), self.width, 1, 0)
        self.get_info_header().resize(
            1, self.width, self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), 0)

    def on_close(self) -> None:
        Tab.on_close(self)
        if config.get_by_tabname('send_chat_states', self.general_jid):
            self.send_chat_state('gone')

    def matching_names(self) -> list[tuple[int, str]]:
        res = []
        jid = self.jid
        res.append((2, jid.bare))
        res.append((1, jid.user))
        contact = roster[self.name]
        if contact and contact.name:
            res.append((0, contact.name))
        return res


class DynamicConversationTab(ConversationTab):
    """
    A conversation tab associated with one bare JID.  It used to
    support resource locking (as described in XEP-0296), but that was a
    bad idea so it has been removed.
    Only one DynamicConversationTab can be opened for a given jid.
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', jid: JID, initial: SMessage | None = None) -> None:
        self.locked_resource = None
        ConversationTab.__init__(self, core, jid, initial=initial)
        self.jid.resource = None
        self.info_header = windows.DynamicConversationInfoWin()
        self.resize()
        self.update_commands()
        self.update_keys()

    def get_info_header(self) -> windows.DynamicConversationInfoWin:
        return self.info_header

    def get_dest_jid(self) -> JID:
        """
        Returns the bare jid.
        """
        return JID(self.jid.bare)

    def refresh(self) -> None:
        """
        Different from the parent class only for the info_header object.
        """
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)
        display_bar = display_info_win = not self.size.tab_degrade_y

        self.text_win.refresh()
        if display_bar:
            self.upper_bar.refresh(self.jid.bare, roster.contacts.get(self.jid.bare))
        displayed_jid = self.jid.bare
        self.get_info_header().refresh(displayed_jid, roster[self.jid.bare],
                                       self.text_win, self.chatstate,
                                       ConversationTab.additional_information)
        if display_info_win:
            self.info_win.refresh()

        self.refresh_tab_win()
        self.input.refresh()

    def refresh_info_header(self) -> None:
        """
        Different from the parent class only for the info_header object.
        """
        displayed_jid = self.jid.bare
        self.get_info_header().refresh(displayed_jid, roster[self.jid.bare],
                                       self.text_win, self.chatstate,
                                       ConversationTab.additional_information)
        self.input.refresh()


class StaticConversationTab(ConversationTab):
    """
    A conversation tab associated with one Full JID. It cannot be locked to
    an different resource or unlocked.
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', jid: JID, initial: SMessage | None = None) -> None:
        ConversationTab.__init__(self, core, jid, initial=initial)
        assert jid.resource
        self.info_header = windows.ConversationInfoWin()
        self.resize()
        self.update_commands()
        self.update_keys()

    async def init_logs(self, initial: SMessage | None = None) -> None:
        # Disable local logs because…
        pass

    def get_info_header(self) -> windows.ConversationInfoWin:
        return self.info_header



"""
Defines the Fifo class

This fifo allows simple communication between a remote poezio
and a local computer, with ssh+cat.
"""

import logging
import os
import threading
from typing import IO
from pathlib import Path

log = logging.getLogger(__name__)


class OpenTrick(threading.Thread):
    """
    A threaded trick to make the open for writing succeed.
    A fifo cannot be opened for writing if it has not been
    yet opened by the other hand for reading.
    So, we just open the fifo for reading and we do not close
    it afterwards, because if the other reader disconnects,
    we will receive a SIGPIPE. And we do not want that.

    (we never read anything from it, obviously)
    """

    fd: IO[str] | None
    path: str | Path

    def __init__(self, path: str | Path) -> None:
        threading.Thread.__init__(self)
        self.path = path
        self.fd = None

    def run(self) -> None:
        self.fd = open(self.path, 'r', encoding='utf-8')


class Fifo:
    """
    Just a simple file handler, writing and reading in a fifo.
    Mode is either 'r' or 'w', just like the mode for the open()
    function.
    """

    fd: IO[str]

    def __init__(self, path: str | Path, mode: str) -> None:
        self.trick = None
        if not os.path.exists(path):
            os.mkfifo(path)
        if mode == 'w':
            self.trick = OpenTrick(path)
            # that thread will wait until we open it for writing
            self.trick.start()
        self.fd = open(path, mode, encoding='utf-8')

    def write(self, data: str) -> None:
        """
        Try to write on the fifo. If that fails, this means
        that nothing has that fifo opened, so the writing is useless,
        so we just return (and display an error telling that, somewhere).
        """
        self.fd.write(data)
        self.fd.flush()

    def readline(self) -> str:
        "Read a line from the fifo"
        return self.fd.readline()

    def __del__(self) -> None:
        "Close opened fds"
        try:
            self.fd.close()
            if self.trick and self.trick.fd:
                self.trick.fd.close()
        except Exception:
            log.exception('Unable to close descriptors for the fifo')



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Define the user class.
A user is a MUC participant, not a roster contact (see contact.py)
"""

import logging
from datetime import timedelta, datetime
from hashlib import md5

from poezio import xhtml, colors
from poezio.common import DissectedPresence
from poezio.theming import get_theme, TextStyle
from slixmpp import JID

log = logging.getLogger(__name__)

ROLE_DICT = {'': 0, 'none': 0, 'visitor': 1, 'participant': 2, 'moderator': 3}


class User:
    """
    keep track of a user in a Room
    """
    __slots__ = ('last_talked', 'jid', 'chatstate', 'affiliation', 'show',
                 'status', 'role', 'nick', 'color', 'occupant_id')
    last_talked: datetime
    color: TextStyle

    def __init__(self,
                 nick: str,
                 affiliation: str,
                 show: str,
                 status: str,
                 role: str,
                 jid: JID | None,
                 color: str = '',
                 *,
                 occupant_id: str | None = None) -> None:
        # The oldest possible time
        self.last_talked: datetime = datetime(1, 1, 1)
        self.update(affiliation, show, status, role)
        self.change_nick(nick)
        self.jid = jid
        self.chatstate: str | None = None
        self.color = (1, 1)
        self.occupant_id = occupant_id
        if color != '':
            self.change_color(color)
        else:
            self.set_deterministic_color()

    @staticmethod
    def from_presence_and_color(pres: DissectedPresence, color: str) -> 'User':
        """Build a user directly from a muc presence"""
        return User(
            nick=pres.from_nick,
            affiliation=pres.affiliation,
            role=pres.role,
            status=pres.status,
            show=pres.show,
            jid=pres.jid,
            occupant_id=pres.occupant_id,
            color=color,
        )

    def set_deterministic_color(self) -> None:
        theme = get_theme()
        if theme.ccg_palette:
            # use XEP-0392 CCG
            if self.jid and self.jid.domain:
                input_ = self.jid.bare
            else:
                input_ = self.nick
            fg_color = colors.ccg_text_to_color(theme.ccg_palette, input_)
            self.color = fg_color, -1
        else:
            mod = len(theme.LIST_COLOR_NICKNAMES)
            nick_pos = int(md5(self.nick.encode('utf-8')).hexdigest(),
                           16) % mod
            self.color = theme.LIST_COLOR_NICKNAMES[nick_pos]

    def update(self, affiliation: str, show: str, status: str, role: str) -> None:
        self.affiliation = affiliation
        self.show = show
        self.status = status
        if role not in ROLE_DICT:  # avoid invalid roles
            role = ''
        self.role = role

    def change_nick(self, nick: str) -> None:
        self.nick = nick

    def change_color(self, color_name: str | None) -> None:
        color = xhtml.colors.get(color_name or '')
        if color is None:
            self.set_deterministic_color()
        else:
            self.color = (color, -1)

    def set_last_talked(self, time: datetime) -> None:
        """
        time: datetime object
        """
        if time > self.last_talked:
            self.last_talked = time

    def has_talked_since(self, t: int) -> bool:
        """
        t: int
        Return True if the user talked since the last s seconds
        """
        delta = timedelta(0, t)
        if datetime.now() - delta > self.last_talked:
            return False
        return True

    def __repr__(self) -> str:
        return ">%s<" % (self.nick)

    def __eq__(self, b: 'User | object') -> bool:
        if not isinstance(b, User):
            return False
        if self.occupant_id:
            return self.occupant_id == b.occupant_id
        else:
            return self is b

    def __gt__(self, b: 'User') -> bool:
        if ROLE_DICT[self.role] == ROLE_DICT[b.role]:
            return self.nick.lower() > b.nick.lower()
        return ROLE_DICT[self.role] < ROLE_DICT[b.role]

    def __ge__(self, b: 'User') -> bool:
        if ROLE_DICT[self.role] == ROLE_DICT[b.role]:
            return self.nick.lower() >= b.nick.lower()
        return ROLE_DICT[self.role] <= ROLE_DICT[b.role]

    def __lt__(self, b: 'User') -> bool:
        if ROLE_DICT[self.role] == ROLE_DICT[b.role]:
            return self.nick.lower() < b.nick.lower()
        return ROLE_DICT[self.role] > ROLE_DICT[b.role]

    def __le__(self, b: 'User') -> bool:
        if ROLE_DICT[self.role] == ROLE_DICT[b.role]:
            return self.nick.lower() <= b.nick.lower()
        return ROLE_DICT[self.role] >= ROLE_DICT[b.role]



import poezio.theming

class IrssiTheme(poezio.theming.Theme):
    COLOR_INFORMATION_BAR = (6, 4)

    COLOR_TAB_NEW_MESSAGE = (15, 4)
    COLOR_TAB_CURRENT = (7, 12)
    COLOR_TAB_HIGHLIGHT = (13, 4)
    COLOR_TAB_DISCONNECTED = (9, 4)

    COLOR_HIGHLIGHT_NICK = (11, -1, 'b')
    COLOR_ME_MESSAGE = (15, -1)

    def __init__(self):
        super().__init__()
        self.LIST_COLOR_NICKNAMES = self.LIST_COLOR_NICKNAMES[:]
        self.LIST_COLOR_NICKNAMES.remove((19, -1))
        self.LIST_COLOR_NICKNAMES.remove((20, -1))
        self.LIST_COLOR_NICKNAMES.remove((226, -1))
        self.LIST_COLOR_NICKNAMES.remove((227, -1))

theme = IrssiTheme()





"""
Classes used to replace the input in some tabs or special situations,
but which are not inputs.
"""

import logging
from poezio.windows.inputs import InputLike
from poezio.theming import get_theme, to_curses_attr

log = logging.getLogger(__name__)


class NoInput(InputLike):
    """A class for when there is no input field on a tab."""
    def refresh(self) -> None:
        return


class HelpText(InputLike):
    """
    A Window just displaying a read-only message.
    Usually used to replace an Input when the tab is in
    command mode.
    """

    def __init__(self, text: str = '') -> None:
        super().__init__()
        self.txt = text

    def refresh(self) -> None:
        """Refresh the window."""
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        self.addstr(0, 0, self.txt[:self.width - 1],
                    to_curses_attr(get_theme().COLOR_INFORMATION_BAR))
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def refresh_txt(self, txt: str | None) -> None:
        """Refresh the window, but with new text."""
        if txt is not None:
            self.txt = txt
        self.refresh()



"""
Display an image URL as an embedded image in some clients like Conversations.
Uses: https://xmpp.org/extensions/xep-0066.html#x-oob

Usage
-----

.. glossary::

    /embed <image_url>

        Run this command to send the <image_url> as an
        embedded image in your contact's client.
"""

from poezio import tabs
from poezio.plugin import BasePlugin
from poezio.theming import get_theme
from poezio.ui.types import Message


class Plugin(BasePlugin):
    def init(self) -> None:
        for tab_t in [tabs.MucTab, tabs.StaticConversationTab, tabs.DynamicConversationTab, tabs.PrivateTab]:
            self.api.add_tab_command(
                tab_t,
                'embed',
                self.embed_image_url,
                help='Embed an image url into the contact\'s client',
                usage='<image_url>')

    def embed_image_url(self, url: str, tab: tabs.Tab | None = None) -> None:
        tab = tab or self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        message = self.core.xmpp.make_message(tab.jid)
        message['body'] = url
        message['oob']['url'] = url
        message['type'] = 'groupchat'
        if not isinstance(tab, tabs.MucTab):
            message['type'] = 'chat'
            tab.add_message(
                Message(
                    message['body'],
                    nickname=tab.core.own_nick,
                    nick_color=get_theme().COLOR_OWN_NICK,
                    identifier=message['id'],
                    jid=tab.core.xmpp.boundjid,
                ),
            )
        message.send()
        self.core.refresh_window()



"""
Module defining the global info bar

This window is the one listing the current opened tabs in poezio.
The GlobalInfoBar can be either horizontal or vertical
(VerticalGlobalInfoBar).
"""
import curses
import itertools
import logging

from typing import List, Optional, TYPE_CHECKING

from poezio.config import config
from poezio.windows.base_wins import Win
from poezio.theming import get_theme, to_curses_attr, TextStyle
from poezio.common import unique_prefix_of
from poezio.colors import ccg_text_to_color

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class GlobalInfoBar(Win):
    __slots__ = ('core')

    def __init__(self, core: 'Core') -> None:
        Win.__init__(self)
        self.core = core

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        theme = get_theme()
        self.addstr(0, 0, "[",
                    to_curses_attr(theme.COLOR_INFORMATION_BAR))

        show_names = config.getbool('show_tab_names')
        show_nums = config.getbool('show_tab_numbers')
        use_nicks = config.getbool('use_tab_nicks')
        show_inactive = config.getbool('show_inactive_tabs')
        unique_prefix_tab_names = config.getbool('unique_prefix_tab_names')
        autocolor_tab_names = config.getbool('autocolor_tab_names')

        if unique_prefix_tab_names:
            unique_prefixes: List[Optional[str]] = [None] * len(self.core.tabs)
            sorted_tab_indices = sorted(
                (str(tab.name), i)
                for i, tab in enumerate(self.core.tabs)
            )
            prev_name = ""
            for (name, i), next_item in itertools.zip_longest(
                    sorted_tab_indices, sorted_tab_indices[1:]):
                # TODO: should this maybe use something smarter than .lower()?
                # something something stringprep?
                name = name.lower()
                prefix_prev = unique_prefix_of(name, prev_name)
                if next_item is not None:
                    prefix_next = unique_prefix_of(name, next_item[0].lower())
                else:
                    prefix_next = name[0]

                # to be unique, we have to use the longest prefix
                if len(prefix_next) > len(prefix_prev):
                    prefix = prefix_next
                else:
                    prefix = prefix_prev

                unique_prefixes[i] = prefix
                prev_name = name

        for nb, tab in enumerate(self.core.tabs):
            if not tab:
                continue
            color: TextStyle = tab.color
            if not show_inactive and color is theme.COLOR_TAB_NORMAL and (
                    tab.priority < 0):
                continue
            if autocolor_tab_names:
                # TODO: in case of private MUC conversations, we should try to
                # get hold of more information to make the colour the same as
                # the nickname colour in the MUC.
                fgcolor, bgcolor, *flags_list = color

                flags = ''.join(flags_list)

                # this is fugly, but I’m not sure how to improve it... since
                # apparently the state is only kept in the color -.-
                if (color == theme.COLOR_TAB_HIGHLIGHT or
                        color == theme.COLOR_TAB_PRIVATE):
                    fgcolor = ccg_text_to_color(theme.ccg_palette, tab.name)
                    bgcolor = -1
                    flags = theme.MODE_TAB_IMPORTANT
                elif color == theme.COLOR_TAB_NEW_MESSAGE:
                    fgcolor = ccg_text_to_color(theme.ccg_palette, tab.name)
                    bgcolor = -1
                    flags = theme.MODE_TAB_NORMAL

                color = (fgcolor, bgcolor, flags)
            try:
                if show_nums or not show_names:
                    self.addstr("%s" % str(nb), to_curses_attr(color))
                    if show_names:
                        self.addstr(' ', to_curses_attr(color))
                if show_names:
                    if unique_prefix_tab_names:
                        self.addstr(unique_prefixes[nb], to_curses_attr(color))
                    elif use_nicks:
                        self.addstr("%s" % str(tab.get_nick()),
                                    to_curses_attr(color))
                    else:
                        self.addstr("%s" % tab.name, to_curses_attr(color))
                self.addstr("|",
                            to_curses_attr(theme.COLOR_INFORMATION_BAR))
            except Exception:  # end of line
                break
        (y, x) = self._win.getyx()
        self.addstr(y, x - 1, '] ',
                    to_curses_attr(theme.COLOR_INFORMATION_BAR))
        (y, x) = self._win.getyx()
        remaining_size = self.width - x
        self.addnstr(' ' * remaining_size, remaining_size,
                     to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self._refresh()


class VerticalGlobalInfoBar(Win):
    __slots__ = ('core')

    def __init__(self, core: 'Core', scr: curses.window) -> None:
        Win.__init__(self)
        self.core = core
        self._win = scr

    def refresh(self) -> None:
        height, width = self._win.getmaxyx()
        self._win.erase()
        sorted_tabs = [tab for tab in self.core.tabs if tab]
        theme = get_theme()
        if not config.getbool('show_inactive_tabs'):
            sorted_tabs = [
                tab for tab in sorted_tabs
                if (
                    tab.vertical_color != theme.COLOR_VERTICAL_TAB_NORMAL or
                    tab.priority > 0
                )
            ]
        nb_tabs = len(sorted_tabs)
        use_nicks = config.getbool('use_tab_nicks')
        if nb_tabs >= height:
            # TODO: As sorted_tabs filters out gap tabs this ensures pos is
            # always set, preventing UnboundLocalError. Now is this how this
            # should be fixed.
            pos = 0
            for y, tab in enumerate(sorted_tabs):
                if tab.vertical_color == theme.COLOR_VERTICAL_TAB_CURRENT:
                    pos = y
                    break
            # center the current tab as much as possible
            if pos < height // 2:
                sorted_tabs = sorted_tabs[:height]
            elif nb_tabs - pos <= height // 2:
                sorted_tabs = sorted_tabs[-height:]
            else:
                sorted_tabs = sorted_tabs[pos - height // 2:pos + height // 2]
        asc_sort = (config.getstr('vertical_tab_list_sort') == 'asc')
        for y, tab in enumerate(sorted_tabs):
            color = tab.vertical_color
            if asc_sort:
                y = height - y - 1
            self.addstr(y, 0, "%2d" % tab.nb,
                        to_curses_attr(theme.COLOR_VERTICAL_TAB_NUMBER))
            self.addstr('.')
            if use_nicks:
                self.addnstr("%s" % tab.get_nick(), width - 4,
                             to_curses_attr(color))
            else:
                self.addnstr("%s" % tab.name, width - 4, to_curses_attr(color))
        separator = to_curses_attr(theme.COLOR_VERTICAL_SEPARATOR)
        self._win.attron(separator)
        self._win.vline(0, width - 1, curses.ACS_VLINE, height)
        self._win.attroff(separator)
        self._refresh()



"""
This plugin allows sending message retractions in a room as a moderator.
It requires XEP-0425 to be available on the room.

Command
-------

.. glossary::

    /moderate
        **Usage in a MUC tab:** ``/moderate <message time> [reason]``
"""

from slixmpp.exceptions import IqError, IqTimeout
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.core.structs import Completion
from poezio.windows.inputs import Input
from poezio.ui.types import Message
from poezio.tabs import MucTab


DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%S'


class Plugin(BasePlugin):
    """Message Moderation (XEP-0425) plugin"""

    def init(self) -> None:
        self.core.xmpp.register_plugin('xep_0425')

        self.api.add_tab_command(
            MucTab,
            name='moderate',
            handler=self.command_moderate,
            usage='<message time> [reason]',
            help='Moderate a message using its timestamp (using tab complete)',
            short='Moderate a message using its timestamp',
            completion=self.completion_moderate,
        )

    def completion_moderate(self, the_input: Input) -> Completion | None:
        """Datetime completion"""
        if the_input.get_argument_position() != 1:
            return None

        all_messages = self.api.get_conversation_messages()[:-100:-1]
        filtered_messages = filter(
            lambda m: isinstance(m, Message) and m.nickname and m.stanza_id,
            all_messages,
        )
        messages = list(map(lambda m: m.time.strftime(DATETIME_FORMAT), filtered_messages))

        if not messages:
            return None

        return Completion(the_input.auto_completion, messages, '')

    @command_args_parser.quoted(1, 1)
    async def command_moderate(self, args: list[str]) -> None:
        """Moderate a message in a chatroom"""
        if not args:
            return
        time, *reason_list = args
        reason = ' '.join(reason_list)

        found_msg = self.api.find_message(time, DATETIME_FORMAT)
        if not found_msg:
            self.api.information(
                f'Moderated message with timestamp “{time}” not found.',
                'Error',
            )
            return

        id_ = found_msg.stanza_id
        if id_ is None:
            self.api.information(
                f'Moderated message with timestamp “{time}” '
                'found but contained no “{urn:xmpp:sid:0}stanza-id” identifier.',
                'Error',
            )
            return

        # For logging
        nickname = found_msg.nickname
        namespace = self.core.xmpp['xep_0425'].namespace

        tab = self.api.current_tab()
        if not isinstance(tab, MucTab):
            return
        jid = tab.jid
        try:
            await self.core.xmpp['xep_0425'].moderate(jid, id_, reason)
        except IqError as exn:
            if exn.iq['error']['text']:
                text = f"\n{exn.iq['error']['text']}"
            else:
                text = ''

            if exn.iq['error']['condition'] == 'service-unavailable':
                self.api.information(
                    f'Room “{jid}” doesn\'t support “{namespace}”.{text}',
                    'Error',
                )
            elif exn.iq['error']['condition'] == 'item-not-found':
                self.api.information(
                    f'Room “{jid}”: Message to moderate not found.{text}',
                    'Error',
                )
            elif exn.iq['error']['condition'] == 'forbidden':
                self.api.information(
                    f'Room “{jid}”: Message moderation forbidden.{text}',
                    'Error',
                )

            return
        except IqTimeout:
            self.api.information(f'Received no reply querying “{jid}”…', 'Error')
            return

        self.api.information(
            f'Room “{jid}”: message from: “{nickname}” at “{time}” successfully retracted.',
            'Info',
        )
